/* (C) 2016-2023 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {Pipe, PipeTransform} from "@angular/core";
import {DomSanitizer, SafeHtml} from "@angular/platform-browser";
// @ts-ignore
import moment from "moment";
import {HeaterConfig} from "./state/PrinterState";

// const moment = momentImported;

@Pipe({name: "byte"})
export class Byte implements PipeTransform {
  transform(input: number): string {
    if (input < 1024) {
      return input + " <?php _('byte') ?>";
    }
    input /= 1024.0;
    if (input < 1024) {
      return input.toFixed(1) + " <?php _('kB') ?>";
    }
    input /= 1024.0;
    if (input < 1024) {
      return input.toFixed(1) + " <?php _('MB') ?>";
    }
    input /= 1024.0;
    return input.toFixed(1) + " <?php _('GB') ?>";
  }
}

@Pipe({name: "yesno"})
export class YesNo implements PipeTransform {
  transform(input: number | boolean): string {
    if (input) {
      return "<?php _('Yes') ?>";
    }
    return "<?php _('No') ?>";
  }
}

@Pipe({name: "coord"})
export class Coord implements PipeTransform {
  transform(input: number, precision: number): string {
    if (input === null || input === undefined) {
      return "------." + "--------".substr(0, precision);
    }
    let r = input.toFixed(precision);
    let extra = 6 + precision - r.length;
    while (extra > 0) {
      extra--;
      r = "&#160;" + r;
    }
    return r;
  }
}

/** Format a float with fixed precision of fractional digits. */
@Pipe({name: "fixed"})
export class Fixed implements PipeTransform {
  transform(input: number, precision: number): string {
    if (input === null || input === undefined) {
      return "-." + "--------".substr(0, precision);
    }
    return input.toFixed(precision);
  }
}

/**
 * Convert a string containing a integer to an integer.
 */
@Pipe({name: "toint"})
export class ToInt implements PipeTransform {
  transform(input: number): number {
    if (input === null || input === undefined) {
      return 0;
    }
    return parseInt(<any>(input), 10);
  }
}

function twoDigits(a) {
  a = a.toFixed(0);
  if (a.length === 1) {
    return "0" + a;
  }
  return a;
}

@Pipe({name: "hms"})
export class HMS implements PipeTransform {
  transform(input: number): string {
    if (input === undefined) {
      return "";
    }
    if (input < 0) {
      return "<?php _('Computing ...') ?>";
    }
    const hours = Math.floor(input / 3600);
    input -= 3600 * hours;
    const min = Math.floor(input / 60);
    input -= min * 60;
    let t = "";
    if (hours > 0) {
      t += hours + "h ";
    }
    if (min > 0 || hours > 0) {
      t += twoDigits(min) + "m ";
    }
    return t + twoDigits(input) + "s";
  }
}

@Pipe({name: "hmsShort"})
export class HMSShort implements PipeTransform {
  transform(input: number): string {
    if (input === undefined) {
      return "";
    }
    if (input < 0) {
      return "<?php _('Computing ...') ?>";
    }
    const hours = Math.floor(input / 3600);
    input -= 3600 * hours;
    const min = Math.floor(input / 60);
    input -= min * 60;
    let t = "";
    if (hours > 0) {
      t += hours + "h ";
    }
    if (min > 0 || hours > 0) {
      t += twoDigits(min) + "m";
    }
    if (hours > 0) {
      return t;
    }
    return t + " " + twoDigits(input) + "s";
  }
}

@Pipe({name: "timeShort"})
export class TimeShort implements PipeTransform {
  transform(input: number): string {
    if (input === undefined) {
      return "";
    }
    const d = moment(input);
    const now = moment();
    return (d.dayOfYear() !== now.dayOfYear() ? d.format("L") + " " : "") + d.format("LT");
  }
}

@Pipe({name: "dateShort"})
export class DateShort implements PipeTransform {
  transform(input: number): string {
    if (input === undefined) {
      return "";
    }
    const d = moment(new Date(input));
    return d.format("L");
  }
}

@Pipe({name: "dateTimeShort"})
export class DateTimeShort implements PipeTransform {
  transform(input: number): string {
    if (input === undefined) {
      return "";
    }
    const d = moment(input);
    return d.format("L") + " " + d.format("LT");
  }
}

@Pipe({name: "dhms"})
export class DHMS implements PipeTransform {
  transform(input: number): string {
    if (input === undefined) {
      return "";
    }
    if (input < 0) {
      return "<?php _('Computing ...') ?>";
    }
    let hours = Math.floor(input / 3600);
    input -= 3600 * hours;
    let days = 0;
    if (hours > 23) {
      days = Math.floor(hours / 24);
      hours -= days * 24;
    }
    const min = Math.floor(input / 60);
    input -= min * 60;
    let t = "";
    if (days > 0) {
      t += days + "d ";
    }
    if (hours > 0 || days > 0) {
      t += hours + "h ";
    }
    if (min > 0 || hours > 0) {
      t += twoDigits(min) + "m ";
    }
    return t + twoDigits(input) + "s";
  }
}

@Pipe({name: "equals"})
export class EqualsPipe implements PipeTransform {
  transform(value: any, name, filt): any {
    if (value === undefined) {
      return value;
    }
    return value.filter(item => item[name] === filt);
  }
}

@Pipe({
  name: "sanitizeHtml"
})
export class SanitizeHtml implements PipeTransform {

  constructor(private _sanitizer: DomSanitizer) {
  }

  transform(v: string): SafeHtml {
    return this._sanitizer.bypassSecurityTrustHtml(v);
  }
}

@Pipe({
  name: "tempRead"
})
export class TempReadPipe implements PipeTransform {
  transform(val: any, digits: number) {
    if (!digits) {
      digits = 0;
    }
    if (typeof val === "undefined") {
      return "";
    }
    if (val === -333) {
      return "<?php _('DEF') ?>";
    }
    if (val === -444) {
      return "<?php _('DEC') ?>";
    }
    return val.toFixed(digits) + "°C";
  }
}

@Pipe({
  name: "tempSet"
})
export class TempOffPipe implements PipeTransform {
  transform(val: any, digits: number) {
    if (!digits) {
      digits = 0;
    }
    if (typeof val === "undefined") {
      return "";
    }
    if (val === 0) {
      return "<?php _('Off') ?>";
    }
    return val.toFixed(digits) + "°C";
  }
}


@Pipe({
  name: "tempFull"
})
export class TempFullPipe implements PipeTransform {
  transform(val: any, digits: number) {
    if (!digits) {
      digits = 0;
    }
    if (!val) {
      return "";
    }
    if (val.currentTemp === -333) {
      return "<?php _('DEF') ?>";
    }
    if (val.currentTemp === -444) {
      return "<?php _('DEC') ?>";
    }
    return val.currentTemp.toFixed(digits) + " / " + (val.tempSet === 0 ? "<?php _('Off') ?>" : val.tempSet.toFixed(0) + "°C");
  }
}

@Pipe({
  name: "tempFullConf"
})
export class TempFullConfPipe implements PipeTransform {
  transform(val: any, conf: HeaterConfig, digits: number) {
    if (!digits) {
      digits = 0;
    }
    if (!val) {
      return "";
    }
    if (val.currentTemp === -333) {
      return "<?php _('DEF') ?>";
    }
    if (val.currentTemp === -444) {
      return "<?php _('DEC') ?>";
    }
    const unit = conf.unit ?? "°C";
    if (conf.supportTemperature) {
      return val.currentTemp.toFixed(digits) + " / " + (val.tempSet === 0 ? "<?php _('Off') ?>" : val.tempSet.toFixed(0) + unit);
    } else {
      return val.currentTemp.toFixed(digits) + unit;
    }
  }
}
@Pipe({
  name: "quickCommands"
})
export class QuickCommandsPipe implements PipeTransform {
  transform(qc: any, printing: boolean) {
    if (!qc) {
      return [];
    }
    return qc.filter((quick) => !(!quick.visibleWhenPrinting && printing));
  }
}
