/* (C) 2016-2024 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {Injectable, ApplicationRef} from "@angular/core";
import {HttpClient} from "@angular/common/http";
import {SubscriptionLike} from "rxjs";
import {Store} from "@ngrx/store";
import type {
  UserState,
  TemperatureDataState,
  RepetierState,
  ConnectionState,
  PrintersState,
  ServerState,
  Language, ExternalCommand
} from "./state/RepetierState";
import {Router} from "@angular/router";
import {ConnectionAction} from "./actions/ConnectionAction";
import {ServerAction} from "./actions/ServerAction";
import {LicenseAction} from "./actions/LicenseAction";
import {ConnectionHandler, ServerEvent} from "./ConnectionHandler";
import {
  PrinterAction,
  PrinterAddLogLineAction,
  PrinterAppendEepromAction,
  PrinterClearEepromAction,
  PrinterSetRecoverAction
} from "./actions/PrinterAction";
import {PrintersAction, PrintersSetStateTimeAction} from "./actions/PrintersAction";
import {UserAction} from "./actions/UserAction";
import {WifiAction} from "./actions/WifiAction";
import {VisualAids} from "./VisualAids";
import {RSUtils} from "./RSUtils";

@Injectable()
export class RepetierServer {
  private disposed = false;
  server: string = window.location.host;
  private conDialogVisible = false;
  private autoUpdaterRunning = false;
  alert = false;
  conDialog = null;
  private firstConnection = true;
  private eventSubscription: SubscriptionLike;
  private connectionEventSubscription: SubscriptionLike;
  private localEventSubscription: SubscriptionLike;
  private connectionState: ConnectionState;
  private userState: UserState;
  private serverState: ServerState;
  private printersState: PrintersState;
  private periodical5000Timer: any;
  private periodical2000Timer: any;

  constructor(public app: ApplicationRef,
              private http: HttpClient,
              private redux: Store<RepetierState>,
              private connectionAction: ConnectionAction,
              private serverAction: ServerAction,
              private licenseAction: LicenseAction,
              private connectionHandler: ConnectionHandler,
              private printersAction: PrintersAction,
              private printerAction: PrinterAction,
              private userAction: UserAction,
              private wifiAction: WifiAction,
              private visualAid: VisualAids,
              private router: Router) {

    (<any>window).rs = this;
    this.http.get<Language[]>("/languages.json").subscribe((data) => {
      this.serverAction.setLanguages(data);
    });
    this.redux.select("connection").subscribe(c => {
      this.connectionState = c;
    });
    this.redux.select("printers").subscribe(c => {
      this.printersState = c;
    });
    this.redux.select("user").subscribe(c => {
      this.userState = c;
    });
    this.redux.select("server").subscribe(c => {
      this.serverState = c;
    });
    this.eventSubscription = this.connectionHandler.serverEvents.subscribe(this.eventListener);
    this.localEventSubscription = this.connectionHandler.localEvents.subscribe(this.localEventListener);
    this.connectionEventSubscription = this.connectionHandler.connectionEvents.subscribe(this.connectionEventListener);
    this.periodical5000Timer = setInterval(this.periodical5000, 5000);
    this.periodical2000Timer = setInterval(this.periodical2000, 2000);
  }

  periodical5000 = () => {
    if (this.connectionState.connected) {
      this.printersAction.updatePrinters();
    }
  };

  periodical2000 = () => {
    if (this.connectionState.connected) {
      this.printersAction.updateStates();
    }
  };

  connectionEventListener = (evt: ServerEvent) => {
    switch (evt.event) {
      case "updateImportant":
        this.updateImportant();
        break;
    }
  };

  localEventListener = (evt: ServerEvent) => {
    switch (evt.event) {
      case "connected":
        if (this.autoUpdaterRunning) { // reload after server update!
          // noinspection JSDeprecatedSymbols
          location.reload();
        }
        this.autoUpdaterRunning = false;
        this.firstConnection = false;
        // Read external commands
        this.connectionHandler.send("listExternalCommands").then((data) => {
          this.serverAction.setAllExternalCommands(data);
          this.updateExtCommands(data);
        });
        this.licenseAction.update();
        this.wifiAction.updateRouterList();
        this.serverAction.updateFolders();
        this.serverAction.updateWebCalls();
        this.serverAction.updateGPIO();
        this.serverAction.updateExternalLinks();
        if (this.conDialogVisible) {
          this.visualAid.onConnected();
        }
        this.serverAction.updateDiskUsage();
        this.conDialogVisible = false;
        this.wifiAction.refreshRouterList(); // not needed, wifi calls it anyway at startup
        break;
      case "disconnected":
        if (!this.firstConnection) {
          this.conDialogVisible = true;
          this.visualAid.onDisconnected(this.autoUpdaterRunning);
        }
        break;
    }
  };

  eventListener = (evt: ServerEvent) => {
    try {
      const data = evt.data;
      switch (evt.event) {
        case "settingChanged":
          this.serverAction.setSettings(data);
          /*RSUtils.extend(this.settings, data);
           this._originalSettings = RSUtils.copyObject(this.settings);*/
          break;
        case "userCredentials": {
          this.userAction.setUserData({
            loggedIn: true,
            login: data.login,
            settings: data.settings,
            permissions: data.permissions
          });
          this.updateExtCommands(this.serverState.allExternalCommands);
          localStorage.setItem("session", JSON.stringify({
            user: data.login,
            session: this.connectionState.session
          }));
          this.serverAction.updateFeatures(this.userState);
          this.connectionHandler.fireLocalEvent("userCredentialsStored");
        }
          break;
        case "updatePrinterState":
          if (this.printersState.printers[evt.slug]) {
            const newState = this.printersAction.prepareStateData(evt.slug, data);
            if (RSUtils.newChanges(this.printersState.printers[evt.slug], newState)) {
              this.printerAction.updateState(evt.slug, newState);
              this.redux.dispatch(PrintersSetStateTimeAction({stateTime: new Date().getTime()}));
            }
          }
          break;
        case "foldersChanged":
          this.serverAction.updateFolders();
          break;
        case "wifiChanged":
          this.wifiAction.setWifiList(data);
          break;
        case "progressUpdate":
          this.visualAid.updateProgress(data);
          break;
        case "webCallsChanged":
          this.serverAction.updateWebCalls();
          break;
        case "gpioListChanged":
          this.serverAction.updateGPIO();
          break;
        case "externalLinksChanged":
          this.serverAction.updateExternalLinks();
          break;
        case "gpioPinChanged":
          this.serverAction.updateGPIO();
          break;
        case "hardwareInfo":
          this.serverAction.updateHardwareInfo(data);
          break;
        // printer related events

        case "temp": {// new temperatures in O,S,T,id,t(ime)
          if (this.printersState.stateTime === 0) {
            break;
          } // wait for first flood of temperatures to maintain order
          const temp: TemperatureDataState = PrintersAction.prepareTemperature(data);
          if (data.id >= 1000) {
            if (data.id >= 2000) {
              this.printerAction.addChamberTemperature(evt.slug, data.id - 2000, temp);
            } else {
              this.printerAction.addBedTemperature(evt.slug, data.id - 1000, temp);
            }
          } else if (data.id < this.printersState.printers[evt.slug].extruders.length) {
            this.printerAction.addExtruderTemperature(evt.slug, data.id, temp);
          }
        }
          break;
        case "jobsChanged": // gcode modes changed
          this.printerAction.gcodeFetchModels(evt.slug);
          break;
        case "gcodeInfoUpdated":
          if (data.list === "jobs") {
            this.printerAction.gcodeFetchPrintqueue(evt.slug);
          } else {
            this.printerAction.gcodeFetchModels(evt.slug);
          }
          break;
        case "lastPrintsChanged":
          this.printerAction.updateLastPrints(evt.slug);
          break;
        case "jobFinished":
          break;
        case "jobStarted":
          break;
        case "jobKilled":
          break;
        case "eepromClear":
          this.printerAction.dispatch(evt.printer.slug, PrinterClearEepromAction, data);
          break;
        case "eepromData":
          this.printerAction.dispatch(evt.printer.slug, PrinterAppendEepromAction, data);
          break;
        case "dialogsChanged":
          this.printerAction.updateDialogs(evt.printer.slug);
          break;
        case "recoverChanged":
          this.printerAction.dispatch(evt.printer.slug, PrinterSetRecoverAction, data);
          break;
        case "move":
          this.printerAction.updateState(evt.printer.slug, {
              posX: data.x,
              posY: data.y,
              posZ: data.z,
              posE: data.e,
              posDE: data.de
            }
          );
          break;
        case "log":
          this.printerAction.dispatch(evt.printer.slug, PrinterAddLogLineAction, data);
          break;
        case "printqueueChanged":
          this.printerAction.gcodeFetchPrintqueue(evt.slug);
          break;
        case "state": {
          const mod: any = this.printersAction.prepareStateData(evt.slug, data);
          this.printerAction.updateState(evt.slug, mod);
        }
          break;
        case "config":
          this.printerAction.updateConfig(evt.slug);
          break;
        case "scriptChanged":
          this.printerAction.updateScript(evt.slug, data);
          break;
        case "firmwareChanged":
          break;
        case "printerSettingChanged":
          break;
        case "licenceChanged":
          this.licenseAction.update();
          break;
        case "modelGroupListChanged":
          this.printerAction.updateGroups(evt.slug);
          break;
        // Image has changed, initiate update
        case "newRenderImage": {
          const key = data.slug + "_" + data.list + data.id;
          const urls: { [key: string]: string } = {};
          urls[key + "l"] = "/dyn/render_image?q=" + data.list + "&id=" + data.id + "&slug=" + data.slug + "&v=" + data.version + "&t=l&tm=" + new Date().getTime() + "&sess=" + encodeURIComponent(this.connectionState.session);
          urls[key + "m"] = "/dyn/render_image?q=" + data.list + "&id=" + data.id + "&slug=" + data.slug + "&v=" + data.version + "&t=m&tm=" + new Date().getTime() + "&sess=" + encodeURIComponent(this.connectionState.session);
          urls[key + "s"] = "/dyn/render_image?q=" + data.list + "&id=" + data.id + "&slug=" + data.slug + "&v=" + data.version + "&t=s&tm=" + new Date().getTime() + "&sess=" + encodeURIComponent(this.connectionState.session);
          this.printerAction.updateImageUrls(evt.slug, urls);
          if (data.list === "jobs") {
            this.printerAction.gcodeFetchPrintqueue(evt.slug);
          } else {
            this.printerAction.gcodeFetchModels(evt.slug);
          }
        }
          break;
        case "timer300":
          this.serverAction.updateDiskUsage();
          break;
        case "changeFilamentRequested":
          this.printersAction.setActiveSlug(evt.slug);
          this.router.navigateByUrl("/changefilament");
          break;
        case "pong":
          this.connectionHandler.send("ping", {"source": "Touchscreen-Pong"});
          break;
        case "autoupdateStarted":
          this.autoUpdaterRunning = true;
          break;
      }
    } catch (ex) {
      console.error("RepetierServer.eventListener:", ex);
    }
  };

  dispose() {
    this.disposed = true;
    this.connectionEventSubscription.unsubscribe();
  }

  private updateExtCommands(all: Array<ExternalCommand>) {
    const a = [];
    const user: UserState = this.userState;
    for (const test of all) {
      if (test.permPrint && !user.canPrint) {
        continue;
      }
      if (test.permConfig && !user.canTouch) {
        continue;
      }
      if (test.permAdd && !user.canAddFiles) {
        continue;
      }
      if (test.permDel && !user.canDeleteFiles) {
        continue;
      }
      a.push(test);
    }
    this.serverAction.setExternalCommands(a);
  }

  // preload modal helper in case we are offline, we need this for message
  private updateImportant() {
    this.printersAction.forceFullUpdate = true;
    this.licenseAction.update();
    this.serverAction.loadSettings();
    this.serverAction.updateFolders();
  }

}

