/* (C) 2016-2023 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {Injectable} from "@angular/core";
import type {
  RepetierState,
  PrinterState,
  GroupState,
  ServerState,
  GCodeFileState,
  TemperatureDataState,
  LicenseState,
  ServerFolderState,
  PrintersState,
  ConnectionState,
  LogState, ServerDialog
} from "../state/RepetierState";
import {
  createAction, props, Store
} from "@ngrx/store";
import {take} from "rxjs/operators";
import {ConnectionHandler} from "../ConnectionHandler";
import {VisualAids} from "../VisualAids";
import {Deferred, RSUtils, SerializationHelper} from "../RSUtils";
import {INITIAL_GCODE_FILE_STATE} from "../state/PrinterState";
import {RSBrowseFolder} from "../BrowseFolder";

export interface ScriptName {
  name: string;
  title: string;
}

export const PrinterSetConfigAction = createAction("PRINTER_SET_CONFIG", props<{ slug: string, payload: any }>());
export const PrinterSetWebcamConfigAction = createAction("PRINTER_SET_WEBCAM_CONFIG", props<{ slug: string, payload: any }>());
export const PrinterSetGroupsAction = createAction("PRINTER_SET_GROUPS", props<{ slug: string, payload: any }>());
export const PrinterUpdateStateAction = createAction("PRINTER_UPDATE_STATE", props<{ slug: string, payload: any }>());
export const PrinterupdateStatusAction = createAction("PRINTER_UPDATE_STATUS", props<{ slug: string, payload: any }>());
export const PrinterUpdateGCodeAction = createAction("PRINTER_UPDATE_GCODE", props<{ slug: string, payload: any }>());
export const PrinterSetImgUrlAction = createAction("PRINTER_SET_IMG_URL", props<{ slug: string, payload: any }>());
export const PrinterUpdateImgUrlsAction = createAction("PRINTER_UPDATE_IMG_URLS", props<{ slug: string, payload: any }>());
export const PrinterSetPrintqueueAction = createAction("PRINTER_SET_PRINTQUEUE", props<{ slug: string, payload: any }>());
export const PrinterSetModelsAction = createAction("PRINTER_SET_MODELS", props<{ slug: string, payload: any }>());
export const PrinterSetLastPrintsAction = createAction("PRINTER_SET_LAST_PRINTS", props<{ slug: string, payload: any }>());
export const PrinterUpdateGCodeFilterAction = createAction("PRINTER_UPDATE_GCODE_FILTER", props<{ slug: string, payload: any }>());
export const PrinterAddBedTemperatureAction = createAction("PRINTER_ADD_BED_TEMPERATURE", props<{ slug: string, payload: any }>());
export const PrinterAddChamberTemperatureAction = createAction("PRINTER_ADD_CHAMBER_TEMPERATURE", props<{ slug: string, payload: any }>());
export const PrinterAddExtruderTemperatureAction = createAction("PRINTER_ADD_EXTRUDER_TEMPERATURE", props<{
  slug: string,
  payload: any
}>());
export const PrinterUpdateScriptAction = createAction("UPDATE_SCRIPT", props<{ slug: string, payload: any }>());
export const PrinterUpdateLogStatusAction = createAction("PRINTER_UPDATE_LOG_STATUS", props<{ slug: string, payload: LogState }>());
export const PrinterAddLogLineAction = createAction("PRINTER_ADD_LOG_LINE", props<{ slug: string, payload: any }>());
export const PrinterResetImageReferenceAction = createAction("PRINTER_RESET_IMAGE_REFERENCE", props<{ slug: string, payload: any }>());
export const PrinterClearEepromAction = createAction("PRINTER_CLEAR_EEPROM", props<{ slug: string, payload: any }>());
export const PrinterAppendEepromAction = createAction("PRINTER_APPEND_EEPROM", props<{ slug: string, payload: any }>());
export const PrinterSetNewEepromAction = createAction("PRINTER_SET_NEW_EEPROM_VALUE", props<{ slug: string, payload: any }>());
export const PrinterCopyEeepromValueToOrigAction = createAction("PRINTER_COPY_EEPROM_VALUE_TO_ORIG", props<{
  slug: string,
  payload: any
}>());
export const PrinterSetRecoverAction = createAction("PRINTER_SET_RECOVER", props<{ slug: string, payload: any }>());
export const PrinterSortByAction = createAction("PRINTER_SORT_BY", props<{ slug: string, payload: any }>());
export const PrinterSortModelsAction = createAction("PRINTER_SORT_MODELS", props<{ slug: string, payload: any }>());
export const PrinterSetDialogsAction = createAction("PRINTER_SET_DIALOGS", props<{ slug: string, payload: any }>());

export const PrinterSetVisibleQuickCommandsAction = createAction("PRINTER_SET_VISIBLE_QUICK_COMMANDS", props<{ slug: string, payload: any }>());
export const PrinterSetVisibleWizardsAction = createAction("PRINTER_SET_VISIBLE_WIZARDS", props<{ slug: string, payload: any }>());

type RSDispatchFunction = ({slug, payload}: { slug: string, payload: any }) => any;

@Injectable()
export class PrinterAction {
  static scriptNames: ScriptName[] = [{name: "start", title: "<?php _('Run before job') ?>"},
    {name: "end", title: "<?php _('Run after job') ?>"},
    {name: "kill", title: "<?php _('Run if job is aborted') ?>"},
    {name: "pause", title: "<?php _('Run on pause') ?>"},
    {name: "continue", title: "<?php _('Run on continue') ?>"},
    {name: "connect", title: "<?php _('Run on connect') ?>"},
    {name: "remove_filament", title: "<?php _('Remove filament') ?>"},
    {name: "insert_filament", title: "<?php _('Insert filament') ?>"},
    {name: "insert_filament_finished", title: "<?php _('After insert filament') ?>"},
    {name: "before_snapshot", title: "<?php _('Before snapshot') ?>"},
    {name: "after_snapshot", title: "<?php _('After snapshot') ?>"}
  ];
  static toolNames = ["<?php _('Extruder') ?>", "<?php _('Milling head') ?>", "<?php _('Laser') ?>", "<?php _('Syringe') ?>", "<?php _('Heated Chamber') ?>"];
  static toolIcons = ["rs rs-extruder", "rs rs-mill", "rs rs-laser", "rs rs-syringe", "rs rs-chamber"];
  static toolDiameterName = ["<?php _('Nozzle diameter') ?>", "<?php _('Milling width') ?>", "<?php _('Laser diameter') ?>", "<?php _('Syringe diameter') ?>", "<?php _('Diameter') ?>"];
  static toolChangeIcons = ["rs rs-filamentchange", "rs rs-mill-change", "rs rs-filamentchange", "rs rs-syringe-change"];
  static toolChangeUpIcons = ["rs rs-filament-out", "rs rs-mill-change-up", "rs rs-filament-out", "rs rs-syringe-change-up"];
  static toolChangeDownIcons = ["rs rs-filament-in", "rs rs-mill-change-down", "rs rs-filament-in", "rs rs-syringe-change-down"];
  static toolChangeNames = ["<?php _('Change Filament') ?>", "<?php _('Change Milling Cutter') ?>", "<?php _('Change Filament') ?>", "<?php _('Change Syringe') ?>"];
  static toolChangeInsert = ["<?php _('Insert Filament') ?>", "<?php _('Insert Milling Cutter') ?>", "<?php _('Insert Filament') ?>", "<?php _('Insert Syringe') ?>"];
  static toolChangeRemove = ["<?php _('Remove Filament') ?>", "<?php _('Remove Milling Cutter') ?>", "<?php _('Remove Filament') ?>", "<?php _('Remove Syringe') ?>"];

  private readingFirmware = false;
  private firmwareSettingsPromise: Deferred = null;

  hasFirmwareSettings = false;
  firmwareSettings: any = {};

  constructor(private store: Store<RepetierState>,
              private connectionHandler: ConnectionHandler,
              private visualAid: VisualAids) {
  }

  get printers(): PrintersState {
    let pe: PrintersState;
    this.store.select("printers").pipe(take(1)).subscribe(printers => {
      pe = printers;
    });
    return pe;
  }

  get connection(): ConnectionState {
    let pe: ConnectionState;
    this.store.select("connection").pipe(take(1)).subscribe(connection => {
      pe = connection;
    });
    return pe;
  }

  get server(): ServerState {
    let pe: ServerState;
    this.store.select("server").pipe(take(1)).subscribe(server => {
      pe = server;
    });
    return pe;
  }

  get license(): LicenseState {
    let pe: LicenseState;
    this.store.select("license").pipe(take(1)).subscribe(license => {
      pe = license;
    });
    return pe;
  }

  getPrinterState(slug: string): PrinterState {
    let s: PrinterState = null;
    this.store.select("printers").pipe(take(1)).subscribe(printers => {
      s = printers.printers[slug];
    });
    if (s === null) {
      console.error("No printer state selected", slug);
    }
    return s;
  }

  getPrinterConfig(slug: string): any {
    const p: PrinterState = this.getPrinterState(slug);
    return p ? p.config : undefined;
  }

  dispatch(slug: string, tp: RSDispatchFunction, data: any): void {
    this.store.dispatch(tp({slug, payload: data}));
  }

  updateScript(slug: string, data: any): void {
    this.dispatch(slug, PrinterUpdateScriptAction, data);
  }

  resetImageReference(slug: string): void {
    this.dispatch(slug, PrinterResetImageReferenceAction, {});
  }

  setConfigData(slug: string, data: any): void {
    this.dispatch(slug, PrinterSetConfigAction, {config: data});
  }

  setDialogs(slug: string, data: any): void {
    this.dispatch(slug, PrinterSetDialogsAction, data);
  }

  setGroups(slug: string, groups: { [key: string]: GroupState }, activeGroup: string) {
    this.dispatch(slug, PrinterSetGroupsAction, {groups: groups, activeGroup: activeGroup});
  }

  setActiveGroupId(slug: string, id: string): void {
    this.dispatch(slug, PrinterUpdateStateAction, {state: {activeGroupId: id}});
  }

  updateState(slug: string, data: any): void {
    this.dispatch(slug, PrinterUpdateStateAction, {state: data});
  }

  updateStatus(slug: string, data: any): void {
    this.dispatch(slug, PrinterupdateStatusAction, {slug: slug, status: data});
  }

  updateGCode(data: GCodeFileState): void {
    this.dispatch(data.slug, PrinterUpdateGCodeAction, data);
  }

  updateImageUrls(slug: string, data: any): void {
    this.dispatch(slug, PrinterUpdateImgUrlsAction, data);
  }

  addBedTemperature(slug: string, idx: number, temp: TemperatureDataState): void {
    this.dispatch(slug, PrinterAddBedTemperatureAction, {id: idx, temp: temp});
  }

  addChamberTemperature(slug: string, idx: number, temp: TemperatureDataState): void {
    this.dispatch(slug, PrinterAddChamberTemperatureAction, {id: idx, temp: temp});
  }

  addExtruderTemperature(slug: string, extrId: number, temp: TemperatureDataState): void {
    this.dispatch(slug, PrinterAddExtruderTemperatureAction, {extruderId: extrId, temp: temp});
  }

  // ----- following commands run server queries -----

  updateData(slug: string): void {
    this.updateGroups(slug); // models use groups, so create them first
    this.gcodeFetchModels(slug);
    this.gcodeFetchPrintqueue(slug);
    this.updateLastPrints(slug);
    this.updateConfig(slug);
    this.fetchScripts(slug);
    this.getRecover(slug);
    this.updateDialogs(slug);
  }

  updateConfig(slug: string): void {
    this.connectionHandler.send("getPrinterConfig", {}, slug).then((data) => {
      this.setConfigData(slug, data);
    });
  }

  recoverBadCondition(slug: string): void {
    this.connectionHandler.send("recoverBadCondition", {}, slug);
  }

  updateDialogs(slug: string): void {
    this.connectionHandler.send("getDialogs", {lang: this.connection.language}, slug).then((data) => {
      if (data.ok) {
        this.setDialogs(slug, data.dialogs);
      }
    });
  }

  selectDialogChoice(idx: number, dialog: ServerDialog, inputEntrys: object): Promise<any> {
    return this.connectionHandler.send("closeDialog", {id: dialog.id, selection: idx, inputData: inputEntrys}, dialog.printer);
  }

  saveTimelapseConfig(slug: string, webcams: any): void {
    const webcams2 = JSON.parse(JSON.stringify(webcams));
    this.connectionHandler.send("setPrinterTimelapseConfig", {webcams: webcams2}, slug).then(() => {
      this.dispatch(slug, PrinterSetWebcamConfigAction, {webcams: webcams2});
    });
  }

  storeNewConfig(slug: string, config: any): Promise<any> {
    return this.connectionHandler.send("setPrinterConfig", config, slug);
  }

  updateGroups(slug: string): void {
    this.connectionHandler.send("listModelGroups", {}, slug).then((data: any) => {
      let hasGroup = false;
      let active = this.printers.printers[slug].activeGroupId;
      const groups: { [key: string]: GroupState } = {"*": {id: "*", name: "<?php _('Group: All') ?>"}};
      for (const idx of Object.keys(data.groupNames)) {
        const v = data.groupNames[idx];
        hasGroup = hasGroup || v === active;
        if (v === "#") {
          groups["#"] = ({id: "#", name: "<?php _('Group: Default') ?>"});
        } else {
          groups[v] = ({id: v, name: "<?php _('Group:') ?> " + v});
        }
      }
      if (!hasGroup) {
        active = "#";
      }
      this.setGroups(slug, groups, active);
    });
  }

  // ----- gcode related functions -----

  gcodeInitPricing(gcodeOrig: GCodeFileState): GCodeFileState {
    const gcode: GCodeFileState = Object.assign({}, gcodeOrig);
    const p: any = this.server.settings["pricing"];
    const map = p.map;
    gcode.perExtruder = [];

    function filamentIndex(uid) {
      if (!uid) {
        return 0;
      }
      let idx = 0;
      for (const i of Object.keys(p.filaments)) {
        const f = p.filaments[i];
        if (f.u === uid) {
          idx = parseInt(i, 10);
        }
      }
      return idx;
    }

    gcode.currency = p.currency;
    for (const idx of Object.keys(gcode.volumeUsage)) {
      const vol = gcode.volumeUsage[idx];
      if (vol <= 0) {
        continue;
      }
      gcode.perExtruder.push({
        extruder: parseInt(idx, 10) + 1,
        mm: gcode.extruderUsage[idx],
        volume: vol * 0.001, price: 0,
        filament: p.filaments[filamentIndex(map[idx + "_" + gcode.slug])],
        weight: 0
      });
    }
    return this.gcodeUpdatePricing(gcode);
  }

  gcodeUpdatePricing(gcodeOrig: GCodeFileState): GCodeFileState {
    const gcode: GCodeFileState = Object.assign({}, gcodeOrig);
    const p: any = this.server.settings["pricing"];
    gcode.weightTotal = 0;
    gcode.total = p.perPrint + p.perHour * gcode.printTime / 3600;
    for (const idx of Object.keys(gcode.perExtruder)) {
      const f = gcode.perExtruder[idx];
      f.weight = f.volume * f.filament.w;
      gcode.weightTotal += f.weight;
      f.price = f.weight * f.filament.p / 1000.0;
      gcode.total += f.price;
    }
    return gcode;
  }

  gcodeDelete(gcode: GCodeFileState): Promise<any> {
    if (gcode.list === "models") {
      return this.connectionHandler.send("removeModel", {id: gcode.id});
    } else {
      return this.connectionHandler.send("removeJob", {id: gcode.id});
    }
  }

  gcodePrint(gcode: GCodeFileState): Promise<any> {
    if (gcode.type === 1) {
      return this.connectionHandler.send("startJob", {id: gcode.id}, gcode.slug);
    } else {
      return this.connectionHandler.send("copyModel", {id: gcode.id}, gcode.slug);
    }
  }

  gcodeQueue(gcode: GCodeFileState): Promise<any> {
    return this.connectionHandler.send("copyModel", {id: gcode.id, autostart: false}, gcode.slug);
  }

  gcodeImageUrl(slug: string, queue, id: number, sz: string, version: number) {
    if (typeof (id) === "undefined") {
      return "/mod/server/img/rendering_" + sz + ".png";
    }
    if (version === undefined) {
      console.log("found undef", slug, queue, id);
    }
    const key = slug + "_" + queue + id + sz;
    // if (!this.printers.printers[slug].imageUrls[key]) {
    let url: string = "/dyn/render_image?q=" + queue + "&id=" + id + "&slug=" + slug + "&t=" + sz + "&v=" + version + "&sess=" + encodeURIComponent(this.connection.session); // + "&tm=" + new Date().getTime();
    /*if (queue === "jobs") {
      url += "&tm=" + new Date().getTime();
    }*/
    if (url !== this.printers.printers[slug].imageUrls[key]) {
      this.dispatch(slug, PrinterSetImgUrlAction, {key: key, url: url});
    }
    // }
    if (this.license.supportPreviews) {
      return this.printers.printers[slug].imageUrls[key];
    }
    return "/img/pro_" + sz + ".png";
  }

  gcodeGetLatestVersion(g: GCodeFileState): GCodeFileState {
    if (!g) {
      return undefined;
    }
    if (g.list === "jobs") {
      for (const x of this.printers.printers[g.slug].queue) {
        if (x.id === g.id) {
          return x;
        }
      }
    } else {
      for (const x of this.printers.printers[g.slug].allModels) {
        if (x.id === g.id) {
          return x;
        }
      }
    }
    return undefined;
  }

  gcodeFetchPrintqueue(slug: string): void {
    this.connectionHandler.send("listJobs", {includeRunning: true}, slug).then((r) => {
      const queue: GCodeFileState[] = [];
      for (const val2 of r.data) {
        const val: GCodeFileState = SerializationHelper.convert(RSUtils.copyObject(INITIAL_GCODE_FILE_STATE), val2);

        val.error = val.analysed === 2;
        val.processing = val.analysed === 0;
        val.ok = val.analysed === 1;
        val.list = "jobs";
        val.slug = slug;
        val.imgL = this.gcodeImageUrl(slug, "jobs", val.id, "l", val2.version);
        val.imgM = this.gcodeImageUrl(slug, "jobs", val.id, "m", val2.version);
        val.imgT = this.gcodeImageUrl(slug, "jobs", val.id, "s", val2.version);
        val.type = 1;
        val.printTimeBest = val.lastPrintTime ? val.lastPrintTime : val.printTime;
        queue.push(this.gcodeInitPricing(val));
      }
      this.dispatch(slug, PrinterSetPrintqueueAction, queue);
    });
  }


  gcodeFetchModels(slug: string): void {
    this.connectionHandler.send("listModels", {}, slug).then((r: any) => {
      const allModels = [];
      for (const val2 of r.data) {
        const val: GCodeFileState = SerializationHelper.convert(RSUtils.copyObject(INITIAL_GCODE_FILE_STATE), val2);

        val.error = val.analysed === 2;
        val.processing = val.analysed === 0;
        val.ok = val.analysed === 1;
        val.list = "models";
        val.slug = slug;
        val.imgL = this.gcodeImageUrl(slug, "models", val.id, "l", val2.version);
        val.imgM = this.gcodeImageUrl(slug, "models", val.id, "m", val2.version);
        val.imgT = this.gcodeImageUrl(slug, "models", val.id, "s", val2.version);
        val.type = 2;
        val.printTimeBest = val.lastPrintTime ? val.lastPrintTime : val.printTime;
        allModels.push(this.gcodeInitPricing(val));
      }
      this.dispatch(slug, PrinterSetModelsAction, allModels);
    });
  }

  updateLastPrints(slug: string): void {
    this.connectionHandler.send("listLastPrints", {}, slug).then((r: any) => {
      this.dispatch(slug, PrinterSetLastPrintsAction, r.prints);
    });
  }

  readEEPROM(slug: string) {
    return this.connectionHandler.send("getEeprom", {}, slug);
  }

  async updateVisibleQuickCommands(slug: string) {
    const data = await this.connectionHandler.send("printerVisibleQuickCommands", {}, slug);
    if (data.ok) {
      this.dispatch(slug, PrinterSetVisibleQuickCommandsAction, data.quickCommands);
    }
    return data.quickCommands;
  }

  async updateVisibleWizards(slug: string) {
    const data = await this.connectionHandler.send("printerVisibleWizards", {}, slug);
    if (data.ok) {
      this.dispatch(slug, PrinterSetVisibleWizardsAction, data.wizards);
    }
  }

  saveEeprom(slug: string) {
    const state = this.getPrinterState(slug);
    if (state.eepromErrors) {
      return;
    }
    return this.connectionHandler.send("setEeprom", {eeprom: state.eeprom}, slug).then(() => {
      this.dispatch(slug, PrinterCopyEeepromValueToOrigAction, {});
    }, () => {
    });
  }

  toolSupportsTemperatures(slug: string, idx: number): boolean {
    const config = this.printers.printers[slug].config;
    if (config && config.extruders && idx >= 0 && idx < config.extruders.length) {
      return config.extruders[idx].supportTemperature;
    }
    return true;
  }


  extruderIcon(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (config && config.extruders && idx >= 0 && idx < config.extruders.length) {
      return PrinterAction.toolIcons[config.extruders[idx].toolType];
    }
    return PrinterAction.toolIcons[0];
  }

  extruderChangeIcon(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (!config || !config.extruders) {
      return "";
    }
    if (idx >= 0 && idx < config.extruders.length) {
      return PrinterAction.toolChangeIcons[config.extruders[idx].toolType];
    }
    return PrinterAction.toolChangeIcons[0];
  }

  extruderChangeUpIcon(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (!config || !config.extruders) {
      return "";
    }
    if (idx >= 0 && idx < config.extruders.length) {
      return PrinterAction.toolChangeUpIcons[config.extruders[idx].toolType];
    }
    return PrinterAction.toolChangeUpIcons[0];
  }

  extruderChangeDownIcon(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (!config || !config.extruders) {
      return "";
    }
    if (idx >= 0 && idx < config.extruders.length) {
      return PrinterAction.toolChangeDownIcons[config.extruders[idx].toolType];
    }
    return PrinterAction.toolChangeDownIcons[0];
  }

  isExtruder(slug: string, idx: number): boolean {
    const config = this.printers.printers[slug].config;
    if (idx >= 0 && idx < config.extruders.length) {
      return config.extruders[idx].toolType === 0;
    }
    return false;
  }

  isCNC(slug: string, idx: number): boolean {
    const config = this.printers.printers[slug].config;
    if (idx >= 0 && idx < config.extruders.length) {
      return config.extruders[idx].toolType === 1;
    }
    return false;
  }

  isLaser(slug: string, idx: number): boolean {
    const config = this.printers.printers[slug].config;
    if (idx >= 0 && idx < config.extruders.length) {
      return config.extruders[idx].toolType === 2;
    }
    return false;
  }

  isSyringe(slug: string, idx: number): boolean {
    const config = this.printers.printers[slug].config;
    if (idx >= 0 && idx < config.extruders.length) {
      return config.extruders[idx].toolType === 3;
    }
    return false;
  }

  extruderDiameterName(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (idx >= 0 && idx < config.extruders.length) {
      return PrinterAction.toolDiameterName[config.extruders[idx].toolType];
    }
    return PrinterAction.toolDiameterName[0];
  }


  extruderName(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (config.extruders.length > idx && config.extruders[idx]) {
      const alias = config.extruders[idx].alias;
      if (alias) {
        return alias;
      }
      return PrinterAction.toolNames[config.extruders[idx].toolType] + " " + (idx + 1);
    }
    return PrinterAction.toolNames[0] + " " + (idx + 1);
  }

  chamberName(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (config.heatedChambers.length > idx) {
      const alias = config.heatedChambers[idx].alias;
      if (alias) {
        return alias;
      }
      return "<?php _('Chamber') ?>" + (config.heatedChambers.length > 1 ? " " + (idx + 1) : "");
    }
    return "<?php _('Chamber') ?>" + (config.heatedChambers.length > 1 ? " " + (idx + 1) : "");
  }

  chamberIcon(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (config.heatedChambers.length > idx) {
      switch (config.heatedChambers[idx].sensorType) {
        case 0:
          return "rs rs-bed";
        case 1:
          return "rs rs-chamber";
        case 2:
          return "rs rs-humidity";
        case 3:
          return "rs rs-water-temp";
        case 4:
          return "rs rs-chip";
        case 5:
          return "rs rs-fan";
        case 6:
          return "rs rs-pressure";
      }
    }
    return "rs rs-chamber";
  }

  bedName(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (config.heatedBeds.length > idx && config.heatedBeds[idx]) {
      const alias = config.heatedBeds[idx].alias;
      if (alias) {
        return alias;
      }
      return "<?php _('Bed') ?>" + (config.heatedChambers.length > 1 ? " " + (idx + 1) : "");
    }
    return "<?php _('Bed') ?>" + " " + (idx + 1);
  }

  fanName(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (config.fans && config.fans[idx]) {
      const alias = config.fans[idx].alias;
      if (alias) {
        return alias;
      }
      return "<?php _('Fan') ?>" + (config.fans.length > 1 ? " " + (idx + 1) : "");
    }
    return "<?php _('Fan') ?>" + " " + (idx + 1);
  }

  extruderChangeName(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (config.extruders) {
      return PrinterAction.toolChangeNames[config.extruders[idx].toolType];
    }
    return PrinterAction.toolChangeNames[0];
  }

  extruderChangeInsert(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (config.extruders) {
      return PrinterAction.toolChangeInsert[config.extruders[idx].toolType];
    }
    return PrinterAction.toolChangeInsert[0];
  }

  extruderChangeRemove(slug: string, idx: number): string {
    const config = this.printers.printers[slug].config;
    if (config.extruders) {
      return PrinterAction.toolChangeRemove[config.extruders[idx].toolType];
    }
    return PrinterAction.toolChangeRemove[0];
  }

  importFiles(slug: string, folder: RSBrowseFolder, files: ServerFolderState[], asJob: boolean) {
    const arr = [];
    for (const i of files) {
      arr.push({name: i.name, url: "folder://" + folder.selectedFolder + i.path, job: asJob});
    }
    if (arr.length === 0) {
      return;
    }
    const printer: PrinterState = this.printers.printers[slug];
    this.connectionHandler.send("importURL", {files: arr, folder: printer.activeGroupId}).then((data) => {
      if (data.ok) {
        this.visualAid.bindProgressMsgToId(data.msgId, "Uploading Files");
        this.visualAid.updateProgress({msgId: data.msgId, finished: false, progress: 0, ok: true});
      }
    }, () => {
    });
  }

  motorsOff(slug: string): Promise<any> {
    return this.connectionHandler.send("motorsOff", {}, slug);
  }

  getRecover(slug: string) {
    this.connectionHandler.send("getRecover", {}, slug).then((data) => {
      this.dispatch(slug, PrinterSetRecoverAction, data.recover);
    }, () => {
    });
  }

  updateGCodeQueues(slug: string) {
    this.gcodeFetchModels(slug);
    this.gcodeFetchPrintqueue(slug);
  }

  setConfigFromFirmware(slug: string): void {
    const p: PrinterState = this.getPrinterState(slug);
    const c: any = RSUtils.copyObject(p.config);
    const f: any = p.firmwareSettings;
    if (f.hasOwnProperty("KeepAliveInterval") && f.KeepAliveInterval > 1000) {
      c.connection.serial.communicationTimeout = Math.round(f.KeepAliveInterval / 1000);
    }
    if (f.hasOwnProperty("baudrate")) {
      c.connection.serial.baudrate = f.baudrate;
    }
    if (f.hasOwnProperty("InputBuffer")) {
      c.connection.serial.inputBufferSize = f.InputBuffer;
    }
    if (f.hasOwnProperty("PrintlineCache")) {
      c.movement.movebuffer = f.PrintlineCache;
    }
    if (f.hasOwnProperty("maxXYSpeed")) {
      c.movement.maxXYSpeed = f.maxXYSpeed;
    }
    if (f.hasOwnProperty("xyJerk")) {
      c.movement.xyJerk = f.xyJerk;
    }
    if (f.hasOwnProperty("maxZSpeed")) {
      c.movement.maxZSpeed = f.maxZSpeed;
    }
    if (f.hasOwnProperty("zJerk")) {
      if (f.zJerk > 0) {
        c.movement.zJerk = f.zJerk;
      } else if (f.xyJerk > 0) {
        c.movement.zJerk = f.xyJerk;
      }
    }
    if (f.hasOwnProperty("xyPrintAcceleration")) {
      c.movement.xyPrintAcceleration = f.xyPrintAcceleration;
    }
    if (f.hasOwnProperty("xyTravelAcceleration")) {
      c.movement.xyTravelAcceleration = f.xyTravelAcceleration;
    }
    if (f.hasOwnProperty("zPrintAcceleration")) {
      c.movement.zPrintAcceleration = f.zPrintAcceleration;
    }
    if (f.hasOwnProperty("zTravelAcceleration")) {
      c.movement.zTravelAcceleration = f.zTravelAcceleration;
    }
    if (f.hasOwnProperty("ZLength")) {
      c.movement.zMax = f.ZLength;
    }
    if (f.hasOwnProperty("RetractionSpeed")) {
      c.movement.G10Speed = f.RetractionSpeed;
    }
    if (f.hasOwnProperty("RetractionLength")) {
      c.movement.G10Distance = f.RetractionLength;
    }
    if (f.hasOwnProperty("RetractionLongLength")) {
      c.movement.G10LongDistance = f.RetractionLongLength;
    }
    if (f.hasOwnProperty("RetractionUndoSpeed")) {
      c.movement.G11Speed = f.RetractionUndoSpeed;
    }
    if (f.hasOwnProperty("RetractionUndoExtraLength")) {
      c.movement.G11ExtraDistance = f.RetractionUndoExtraLength;
    }
    if (f.hasOwnProperty("RetractionUndoExtraLongLength")) {
      c.movement.G11ExtraLongDistance = f.RetractionUndoExtraLongLength;
    }
    if (f.hasOwnProperty("RetractionZLift")) {
      c.movement.G10ZLift = f.RetractionZLift;
    }
    if (f.hasOwnProperty("printerType")) {
      c.general.printerVariant = "cartesian";
      if (f.printerType === 2) {
        c.general.printerVariant = "delta";
        c.shape.basicShape.shape = "circle";
        if (f.hasOwnProperty("printableRadius")) {
          c.shape.basicShape.radius = f.printableRadius;
          c.shape.basicShape.xMin = c.movement.xMin = -f.printableRadius;
          c.shape.basicShape.xMax = c.movement.xMax = f.printableRadius;
          c.shape.basicShape.yMin = c.movement.yMin = -f.printableRadius;
          c.shape.basicShape.yMax = c.movement.yMax = f.printableRadius;
        }
      }
      if (f.printerType === 1) {
        if (f.hasOwnProperty("XMin")) {
          c.shape.basicShape.xMin = c.movement.xMin = f.XMin;
          if (f.hasOwnProperty("XLength")) {
            c.shape.basicShape.xMax = c.movement.xMax = f.XMin + f.XLength;
          }
        }
        if (f.hasOwnProperty("YMin")) {
          c.shape.basicShape.yMin = c.movement.yMin = f.YMin;
          if (f.hasOwnProperty("YLength")) {
            c.shape.basicShape.yMax = c.movement.yMax = f.YMin + f.YLength;
          }
        }
        if (f.hasOwnProperty("ZMin")) {
          c.shape.basicShape.zMin = c.movement.zMin = f.ZMin;
          if (f.hasOwnProperty("ZLength")) {
            c.shape.basicShape.zMax = c.movement.zMax = f.ZMin + f.ZLength;
          }
        }
      }
    }
    if (f.hasOwnProperty("hasHeatedBed")) {
      c.heatedBed.installed = Boolean(f.hasHeatedBed);
    }
    if (f.hasOwnProperty("maxBedTemp")) {
      c.heatedBed.maxTemp = f.maxBedTemp;
    }
    if (f.hasOwnProperty("hasHeatedChamber")) {
      c.heatedChamber.installed = Boolean(f.hasHeatedChamber);
    }
    if (f.hasOwnProperty("maxChamberTemp")) {
      c.heatedChamber.maxTemp = f.maxChamberTemp;
    }
    if (f.hasOwnProperty("sdInstalled")) {
      c.general.sdcard = Boolean(f.sdInstalled);
    }
    if (f.hasOwnProperty("fanInstalled")) {
      c.general.fan = Boolean(f.fanInstalled);
    }
    if (f.hasOwnProperty("fan2Installed")) {
      c.general.fan2 = Boolean(f.fan2Installed);
    }
    if (f.hasOwnProperty("softwarePowerSwitch")) {
      c.general.softwarePower = Boolean(f.softwarePowerSwitch);
    }
    if (f.hasOwnProperty("CaseLights")) {
      c.general.softwareLight = Boolean(f.CaseLights);
    }
    for (let idx = 0; idx < c.extrudes.length; idx++) {
      const ex = c.extruders[idx];
      const app = "[" + idx + "]";
      if (f.hasOwnProperty("extrJerk" + app)) {
        ex.eJerk = f["extrJerk" + app];
      }
      if (f.hasOwnProperty("extrMaxSpeed" + app)) {
        ex.maxSpeed = f["extrMaxSpeed" + app];
      }
      if (f.hasOwnProperty("extrMaxTemp" + app)) {
        ex.maxTemp = f["extrMaxTemp" + app];
      }
      if (f.hasOwnProperty("extrAcceleration" + app)) {
        ex.acceleration = f["extrAcceleration" + app];
      }

    }
  }

  catchFirmwareSettings = (slug: string) => {
    this.connectionHandler.send("getFirmwareSettings", {}, slug).then((data) => {
      this.firmwareSettings = data.settings;
      this.hasFirmwareSettings = true;
      this.setConfigFromFirmware(slug);
      this.readingFirmware = false;
      this.connectionHandler.fireLocalEvent("firmwareConfigRead", this.firmwareSettings);
      if (this.firmwareSettingsPromise != null) {
        this.firmwareSettingsPromise.resolve(this.firmwareSettings);
        this.firmwareSettingsPromise = null;
      }
    });
  };

  updateNewPrinter(slug: string) {
    this.updateState(slug, {
      hasFirmwareSettings: false,
      firmwareSettings: []
    });
    this.connectionHandler.send("getFirmwareSettings", {}, slug).then((data: any) => {
      this.updateState(slug, {
        hasFirmwareSettings: true,
        firmwareSettings: data.settings
      });
      this.connectionHandler.fireLocalEvent("firmwareConfigRead", data.settings, this.getPrinterState(slug));
    });
  }

  updateFirmwareAutoset(slug: string) {
    this.connectionHandler.fireLocalEvent("firmwareConfigRead", this.firmwareSettings, this.getPrinterState(slug));
  }

  updateFirmwareSettings(slug: string) {
    this.readingFirmware = true;
    this.firmwareSettingsPromise = new Deferred();
    this.connectionHandler.send("startUpdateFirmwareSettings", {}, slug).then(() => {
      setTimeout(() => this.catchFirmwareSettings(slug), 3000);
    });
    return this.firmwareSettingsPromise.promise;
  }

  setExtruderTemperature(slug: string, extr: number, temp: number): Promise<any> {
    return this.connectionHandler.send("setExtruderTemperature", {temperature: temp, extruder: extr}, slug);
  }

  setBedTemperature(slug: string, idx: number, temp: number): Promise<any> {
    return this.connectionHandler.send("setBedTemperature", {bedId: idx, temperature: temp}, slug);
  }

  setChamberTemperature(slug: string, idx: number, temp: number): Promise<any> {
    return this.connectionHandler.send("setChamberTemperature", {chamberId: idx, temperature: temp}, slug);
  }

  toggleShutdownAfterPrint(slug: string): Promise<any> {
    return this.connectionHandler.send("setShutdownAfterPrint", {shutdown: !this.getPrinterState(slug).shutdownAfterPrint}, slug);
  }

  power(slug: string) {
    return this.connectionHandler.send("send", {cmd: "@runButtonCommand togglePower"}, slug);
  }

  sendGCode(slug: string, cmd: string): Promise<any> {
    return this.connectionHandler.send("send", {cmd: cmd}, slug);
  }

  sendQuickCommand(slug: string, name: string): Promise<any> {
    return this.connectionHandler.send("sendQuickCommand", {name: name}, slug);
  }

  sendWizardCommand(slug: string, name: string): Promise<any> {
    return this.connectionHandler.send("sendWizardCommand", {name: name}, slug);
  }

  emergencyStop(slug: string): Promise<any> {
    return this.connectionHandler.send("emergencyStop", {}, slug);
  }

  setSpeedMultiply(slug: string, diff: number): Promise<any> {
    this.updateState(slug, {speedMultiply: diff});
    return this.connectionHandler.send("setSpeedMultiply", {speed: diff}, slug);
  }

  setFlowMultiply(slug: string, diff: number): Promise<any> {
    this.updateState(slug, {flowMultiply: diff});
    return this.connectionHandler.send("setFlowMultiply", {speed: diff}, slug);
  }

  setFanPercent(slug: string, idx: number, newSpeed: number): Promise<any> {
    this.updateState(slug, {fanPercent: newSpeed});
    return this.connectionHandler.send("setFanSpeed", {fanId: idx, speed: Math.round(newSpeed * 2.55)}, slug);
  }

  setFanOn(slug: string, idx: number, on: boolean): Promise<any> {
    const p: PrinterState = this.getPrinterState(slug);
    let percent: number = p.fans[idx].percent;
    if (percent === 0) {
      percent = 100;
    }
    this.updateState(slug, {fanOn: on, fanPercent: percent});
    return this.connectionHandler.send("setFanSpeed", {fanId: idx, speed: Math.round(percent * 2.55), "on": on}, slug);
  }

  toggleCaseLights(slug: string): Promise<any> {
    if (this.getPrinterState(slug).lightsOn) {
      return this.connectionHandler.send("setCaseLights", {on: false}, slug);
    }
    return this.connectionHandler.send("setCaseLights", {on: true}, slug);
  }

  xMoveRel(slug: string, x: number): Promise<any> {
    if (this.getPrinterConfig(slug).movement.invertX) {
      x = -x;
    }
    return this.connectionHandler.send("move", {x: x, relative: true}, slug);
  }

  yMoveRel(slug: string, y: number): Promise<any> {
    const p: PrinterState = this.getPrinterState(slug);
    if (p.config.movement.invertY) {
      y = -y;
    }
    return this.connectionHandler.send("move", {y: y, relative: true}, slug);
  }

  zMoveRel(slug: string, z: number): Promise<any> {
    const p: PrinterState = this.getPrinterState(slug);
    if (p.config.movement.invertZ) {
      z = -z;
    }
    return this.connectionHandler.send("move", {z: z, relative: true}, slug);
  }

  xyMoveRel(slug: string, x: number, y: number): Promise<any> {
    const p: PrinterState = this.getPrinterState(slug);
    if (p.config.movement.invertX) {
      x = -x;
    }
    if (p.config.movement.invertY) {
      y = -y;
    }
    return this.connectionHandler.send("move", {x: x, y: y, relative: true}, slug);
  }

  eMoveRel(slug: string, x: number, speed: number): Promise<any> {
    return this.connectionHandler.send("move", {e: x, relative: true, speed: speed}, slug);
  }

  selectExtruder(slug: string, extr: number): void {
    this.connectionHandler.send("activateExtruder", {extruder: extr}, slug);
    setTimeout(() => {
      this.connectionHandler.fireLocalEvent("updateStates");
    }, 20);
  }

  activate(slug: string): Promise<any> {
    return this.connectionHandler.send("activate", {printer: slug}, slug);
  }

  deactivate(slug: string): Promise<any> {
    return this.connectionHandler.send("deactivate", {printer: slug}, slug);
  }

  stopPrint(slug: string): void {
    const p: PrinterState = this.getPrinterState(slug);
    if (p.jobid) {
      this.connectionHandler.send("stopJob", {id: p.jobid});
    }
  }

  pausePrint(slug: string): Promise<any> {
    this.updateState(slug, {paused: true});
    return this.connectionHandler.send("send", {cmd: "@pause <?php _('User requested pause.') ?>"}, slug);
  }

  unpausePrint(slug: string): Promise<any> {
    // this.updateState(slug, {paused: false});
    return this.connectionHandler.send("continueJob", {}, slug);
  }

  executeNamedFunction(slug: string, f: string, value?: number): void {
    const p: PrinterState = this.getPrinterState(slug);
    switch (f) {
      case "xLeft":
        this.xMoveRel(slug, -p.xyzStepSize);
        break;
      case "xRight":
        this.xMoveRel(slug, p.xyzStepSize);
        break;
      case "yUp":
        this.yMoveRel(slug, p.xyzStepSize);
        break;
      case "yUpXLeft":
        this.xyMoveRel(slug, -p.xyzStepSize, p.xyzStepSize);
        break;
      case "yUpXRight":
        this.xyMoveRel(slug, p.xyzStepSize, p.xyzStepSize);
        break;
      case "yDown":
        this.yMoveRel(slug, -p.xyzStepSize);
        break;
      case "yDownXLeft":
        this.xyMoveRel(slug, -p.xyzStepSize, -p.xyzStepSize);
        break;
      case "yDownXRight":
        this.xyMoveRel(slug, p.xyzStepSize, -p.xyzStepSize);
        break;
      case "zUp":
        this.zMoveRel(slug, p.xyzStepSize);
        break;
      case "zDown":
        this.zMoveRel(slug, -p.xyzStepSize);
        break;
      case "eExtrude":
        this.eMoveRel(slug, p.eStepSize, p.extrudeSpeed);
        break;
      case "eRetract":
        this.eMoveRel(slug, -p.eStepSize, p.extrudeSpeed);
        break;
      case "eExtrudeFast":
        this.eMoveRel(slug, p.eStepSize, p.retractSpeed);
        break;
      case "eRetractFast":
        this.eMoveRel(slug, -p.eStepSize, p.retractSpeed);
        break;
      case "home":
        this.connectionHandler.send("send", {cmd: "@runButtonCommand homeAll"}, slug);
        break;
      case "homeX":
        this.connectionHandler.send("send", {cmd: "@runButtonCommand homeX"}, slug);
        break;
      case "homeY":
        this.connectionHandler.send("send", {cmd: "@runButtonCommand homeY"}, slug);
        break;
      case "homeZ":
        this.connectionHandler.send("send", {cmd: "@runButtonCommand homeZ"}, slug);
        break;
      case "motorsOff":
        this.motorsOff(slug);
        break;
      case "autolevel":
        this.connectionHandler.send("send", {cmd: "@runButtonCommand autolevel"}, slug);
        break;
    }
  }

  printingImage(slug: string, imgSize: string): string {
    const p: PrinterState = this.getPrinterState(slug);
    if (!p || !p.printing) {
      return "";
    }
    return this.gcodeImageUrl(slug, "jobs", p.jobid, imgSize || "l", 0);
  }

  fetchScripts(slug: string): void {
    for (let i = 0; i < PrinterAction.scriptNames.length; i++) {
      this.connectionHandler.send("getScript", {name: PrinterAction.scriptNames[i].name}, slug).then((s: any) => {
        this.updateScript(slug, s);
      }, () => {
      });
    }
  }

  runScript(slug: string, scriptName: string): Promise<any> {
    return this.connectionHandler.send("runScript", {script: scriptName}, slug);
  }

  preheat(slug: string, ext: number): Promise<any> {
    return this.connectionHandler.send("preheat", {extruder: ext, bed: 21, chamber: 1}, slug);
  }

  cooldown(slug: string): Promise<any> {
    return this.connectionHandler.send("cooldown", {extruder: 1, bed: 1, chamber: 1}, slug);
  }

  babystepZ(slug: string, z: number): Promise<any> {
    return this.connectionHandler.send("babystep", {z: z}, slug);
  }

  getLoglevel(slug: string): number {
    const p: PrinterState = this.getPrinterState(slug);
    return 12 + (p.log.logCommands ? 1 : 0) + (p.log.logACK ? 2 : 0);
  }

  sendLogLevel(slug: string): Promise<any> {
    return this.connectionHandler.send("setLoglevel", {logPrinter: slug, level: this.getLoglevel(slug)});
  }

  setLogCommands(slug: string, on: boolean): Promise<any> {
    this.dispatch(slug, PrinterUpdateLogStatusAction, {logCommands: on});
    return this.sendLogLevel(slug);
  }

  setLogACK(slug: string, on: boolean): Promise<any> {
    this.dispatch(slug, PrinterUpdateLogStatusAction, {logACK: on});
    return this.sendLogLevel(slug);
  }

  setLogM105(slug: string, on: boolean): Promise<any> {
    this.dispatch(slug, PrinterUpdateLogStatusAction, {logM105: on});
    return this.sendLogLevel(slug);
  }

  setLogPause(slug: string, on: boolean): void {
    this.dispatch(slug, PrinterUpdateLogStatusAction, {paused: on});
  }

}
