/* (C) 2016-2023 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {Injectable} from "@angular/core";
import type {
  RepetierState,
  PrintersState,
  PrinterState,
  ToolState,
  TemperatureDataState
} from "../state/RepetierState";
import {
  INITIAL_PRINTER_STATE,
  INITAL_TOOL_STATE
} from "../state/PrinterState";
import {PrinterAction} from "./PrinterAction";
import {VisualAids} from "../VisualAids";
import {RSUtils} from "../RSUtils";
import {
  createAction, props, Store
} from "@ngrx/store";
import {take} from "rxjs/operators";
import {ConnectionHandler, ServerEvent} from "../ConnectionHandler";
import {PrinterErrorAction} from "./PrinterErrorAction";
import {globalOptions} from "./../globalOptions";

export const PrintersSetActiveAction = createAction("PRINTERS_SET_ACTIVE", props<{ slug: string }>());
export const PrintersCreateAction = createAction("PRINTERS_CREATE", props<PrinterState>());
export const PrintersDestroyAction = createAction("PRINTERS_DESTROY", props<{ slug: string }>());
export const PrintersSetAutoSelectAction = createAction("PRINTERS_SET_AUTO_SELECT", props<{ on: boolean }>());
export const PrintersSetStateTimeAction = createAction("PRINTER_SET_STATE_TIME", props<{ stateTime: number }>());

@Injectable()
export class PrintersAction {
  private updatingState = false;
  private updatingStatus = false;
  private sendMovesCounter = 0;
  public forceFullUpdate = false;

  static prepareTemperature(temp: any): TemperatureDataState {
    return {
      output: temp.O,
      time: temp.t,
      tempSet: temp.S,
      temp: temp.T
    };
  }

  private static preparePrinterListData(data): any {
    const out: any = {
      slug: data.slug,
      online: data.online,
      active: data.active,
      name: data.name
    };
    if (data.jobstate !== "running") {
      out.printJob = "";
      out.printing = false;
      out.paused = false;
      out.printDone = 0;
      out.jobid = 0;
    } else {
      out.printJob = data.job;
      out.printing = true;
      out.printDone = data.done;
      out.jobid = data.jobid;
      out.jobstate = data.jobstate;
      out.paused = data.paused;
      out.pauseState = data.pauseState;
      out.ofLayer = data.ofLayer;
      out.printedTimeComp = data.printedTimeComp;
      out.printedTime = data.printTime;
      out.startTime = data.start;
    }
    out.idle = !out.printing || out.paused;
    return out;
  }

  private static prepareToolFromState(tool: ToolState, data: any) {
    if (data.history) {
      for (const temp of data.history) {
        const newT: TemperatureDataState = PrintersAction.prepareTemperature(temp);
        while (tool.tempHistory.length > 0 && tool.tempHistory[0].time + 300000 < newT.time) {
          tool.tempHistory.splice(0, 1);
        }
        tool.tempHistory.push(newT);
      }
    }
    tool.currentTemp = data.tempRead;
    tool.tempSet = data.tempSet;
    tool.output = data.output;
  }

  constructor(private redux: Store<RepetierState>,
              private connectionHandler: ConnectionHandler,
              private visualAid: VisualAids,
              private printerAction: PrinterAction,
              private printerErrorAction: PrinterErrorAction) {
    this.connectionHandler.connectionState.subscribe(connected => {
      if (connected) {
        this.updatingState = false;
        this.updatingStatus = false;
        if (this.sendMovesCounter > 0) {
          this.connectionHandler.send("sendMoves");
        }
        const pList = this.printers.printers;
        for (const p in pList) {
          if (pList.hasOwnProperty(p)) {
            this.printerAction.sendLogLevel(p);
            this.printerAction.resetImageReference(p);
          }
        }
        this.updatePrinters();
      }
    });
    this.connectionHandler.localEvents.subscribe(this.localEventHandler);
  }

  get printers(): PrintersState {
    let pe: PrintersState;
    this.redux.select("printers").pipe(take(1)).subscribe(printers => {
      pe = printers;
    });
    return pe;
  }

  private localEventHandler = (e: ServerEvent) => {
    if (e.event === "updateStates") {
      this.updateStates();
    }
  };

  setActiveSlug(slug: string): void {
    this.redux.dispatch(PrintersSetActiveAction({slug}));
  }

  setAutoSelect(on: boolean): void {
    this.redux.dispatch(PrintersSetAutoSelectAction({on}));
  }

  addPrinter(printer: PrinterState): void {
    this.redux.dispatch(PrintersCreateAction(printer));
  }

  delPrinter(slug: string): void {
    this.redux.dispatch(PrintersDestroyAction({slug}));
  }

  public prepareStateData(slug: string, data: any): any {
    const printer: PrinterState = this.printers.printers[slug];
    if (!printer || printer.config.extruders === undefined) {
      return {};
    }
    if (data.x === undefined) {
      return data;
    }
    const out: any = {};
    out.fans = RSUtils.copyArray(data.fans);
    out.fans.forEach(x => {
      x.percent = Math.floor(0.5 + x.voltage / 2.55);
    });
    out.activeExtruderId = data.activeExtruder;
    out.debugLevel = data.debugLevel;
    out.firmware = data.firmware;
    out.firmwareURL = data.firmwareURL;
    out.speedMultiply = data.speedMultiply;
    out.flowMultiply = data.flowMultiply;
    out.hasXHome = data.hasXHome;
    out.hasYHome = data.hasYHome;
    out.hasZHome = data.hasZHome;
    out.layer = data.layer;
    out.powerOn = data.powerOn;
    out.doorOpen = data.doorOpen;
    out.shutdownAfterPrint = data.shutdownAfterPrint;
    out.sdcardMounted = data.sdcardMounted;
    out.rec = data.rec;
    out.volumetric = data.volumetric;
    out.lightsOn = data.lights;
    out.posX = data.x;
    out.posY = data.y;
    out.posZ = data.z;
    out.extruders = RSUtils.copyArray(printer.extruders);
    out.heatedBeds = RSUtils.copyArray(printer.heatedBeds);
    out.heatedChambers = RSUtils.copyArray(printer.heatedChambers);
    out.condition = data.condition;
    out.conditionReason = data.conditionReason;
    out.perm = data.perm ?? printer.perm;
    out.gperm = data.gperm ?? printer.gperm;
    out.global = data.global ?? printer.global;
    out.sglobal = data.sglobal ?? printer.sglobal;
    let tempError = false;
    if (data.numExtruder < printer.numExtruder) {
      out.extruders.splice(data.numExtruder, out.extruders.length - data.numExtruder);
    }
    while (data.numExtruder > out.extruders.length) {
      const t: ToolState = RSUtils.copyObject(INITAL_TOOL_STATE);
      t.id = out.extruders.length;
      out.extruders.push(t);
    }
    out.numExtruder = data.numExtruder;
    const numBeds = printer.config.heatedBeds.length;
    const numChambers = printer.config.heatedChambers.length;
    if (data.heatedBeds.length < numBeds) {
      out.heatedBeds.splice(numBeds, data.heatedBeds.length - numBeds);
    }
    while (numBeds > out.heatedBeds.length) {
      const t: ToolState = RSUtils.copyObject(INITAL_TOOL_STATE);
      t.id = out.heatedBeds.length + 1000;
      out.heatedBeds.push(t);
    }
    if (data.heatedChambers.length < numChambers) {
      out.heatedChambers.splice(numChambers, data.heatedChambers.length - numChambers);
    }
    while (numChambers > out.heatedChambers.length) {
      const t: ToolState = RSUtils.copyObject(INITAL_TOOL_STATE);
      t.id = out.heatedChambers.length + 2000;
      out.heatedChambers.push(t);
    }
    for (let i = 0; i < out.numExtruder; i++) {
      PrintersAction.prepareToolFromState(out.extruders[i], data.extruder[i]);
      const bid = PrinterErrorAction.idExtruder(i, slug);
      this.printerErrorAction.createIfNeeded(bid, slug);
      tempError = tempError || this.printerErrorAction.updateHeaterError(bid, out.extruders[i].error);
    }
    if (data.heatedBeds) {
      for (let i = 0; i < numBeds; i++) {
        PrintersAction.prepareToolFromState(out.heatedBeds[i], data.heatedBeds[i]);
        const bid = PrinterErrorAction.idBed(i, slug);
        this.printerErrorAction.createIfNeeded(bid, slug);
        tempError = tempError || this.printerErrorAction.updateHeaterError(bid, data.heatedBeds[i].error);
      }
    }
    if (data.heatedChambers) {
      for (let i = 0; i < numChambers; i++) {
        PrintersAction.prepareToolFromState(out.heatedChambers[i], data.heatedChambers[i]);
        const bid = PrinterErrorAction.idChamber(i, slug);
        this.printerErrorAction.createIfNeeded(bid, slug);
        tempError = tempError || this.printerErrorAction.updateHeaterError(bid, data.heatedChambers[i].error);
      }
    }
    if (tempError) {
      if (this.printerErrorAction.hasUnseenErrors()) {
        this.visualAid.onPrinterError();
      } else {
        this.visualAid.onPrinterErrorSolved();
      }
    }
    return out;
  }

  updatePrinters = () => {
    if (this.updatingStatus) {
      return;
    } // no double query
    this.updatingStatus = true;
    this.connectionHandler.send("listPrinter").then((data) => {
      const exists: { [key: string]: boolean } = {};
      const printers: PrintersState = this.printers;
      for (const p in printers.printers) {
        if (printers.printers.hasOwnProperty(p)) {
          exists[p] = false;
        }
      }
      for (const p of data) {
        if (!printers.printers[p.slug]) {
          exists[p.slug] = true;
          this.addPrinter(Object.assign({}, INITIAL_PRINTER_STATE, PrintersAction.preparePrinterListData(p)));
          this.printerAction.sendLogLevel(p.slug);
          this.printerAction.updateData(p.slug);
        } else {
          exists[p.slug] = true;
          const newStatus = PrintersAction.preparePrinterListData(p);
          if (RSUtils.newChanges(printers.printers[p.slug], newStatus)) {
            this.printerAction.updateStatus(p.slug, newStatus);
          }
          if (this.forceFullUpdate) {
            this.printerAction.sendLogLevel(p.slug);
            this.printerAction.updateData(p.slug);
          }
        }
      }
      for (const key in exists) {
        if (!exists.hasOwnProperty(key)) {
          continue;
        }
        if (exists[key] === false) {
          this.delPrinter(key);
        }
      }
      this.updatingStatus = false;
      this.forceFullUpdate = false;
    });
  }

  updateStates = () => {
    if (this.updatingState) {
      return;
    }
    const printers = this.printers;
    this.updatingState = true;
    this.connectionHandler.send("stateList",
      {
        includeHistory: printers.stateTime === 0,
        perm: globalOptions.permInState,
        gperm: globalOptions.gpermInState,
        global: globalOptions.globalInState,
        sglobal: globalOptions.sglobalInState
      }).then((data) => {
      const stateTime = new Date().getTime(); // Time of last state change
      let update = false;
      for (const key in data) {
        if (!data.hasOwnProperty(key)) {
          continue;
        }
        if (printers.printers[key]) {
          const newState = this.prepareStateData(key, data[key]);
          if (RSUtils.newChanges(printers.printers[key], newState)) {
            this.printerAction.updateState(key, newState);
            update = true;
          }
        }
      }
      this.updatingState = false;
      if (update) {
        this.redux.dispatch(PrintersSetStateTimeAction({stateTime}));
      }
    });
  };

  sendMoves() {
    if (this.sendMovesCounter === 0) {
      this.connectionHandler.send("sendMoves");
    }
    this.sendMovesCounter++;
  }

  hideMoves() {
    if (this.sendMovesCounter === 1) {
      this.connectionHandler.send("hideMoves");
    }
    if (this.sendMovesCounter > 0) {
      this.sendMovesCounter--;
    }
  }
}
