/* (C) 2016-2023 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {INITIAL_PRINTERS_STATE} from "../state/PrintersState";
import type {
  PrinterState,
  PrintersState
} from "../state/RepetierState";
import {
  PrintersCreateAction,
  PrintersDestroyAction,
  PrintersSetActiveAction,
  PrintersSetAutoSelectAction, PrintersSetStateTimeAction
} from "../actions/PrintersAction";
import {Action, createReducer, on} from "@ngrx/store";
import {GCodeFileState, GroupState, LogState, TemperatureDataState, ToolState} from "../state/PrinterState";
import {
  PrinterAddBedTemperatureAction,
  PrinterAddChamberTemperatureAction,
  PrinterAddExtruderTemperatureAction,
  PrinterAddLogLineAction,
  PrinterAppendEepromAction,
  PrinterClearEepromAction,
  PrinterCopyEeepromValueToOrigAction,
  PrinterResetImageReferenceAction,
  PrinterSetConfigAction,
  PrinterSetDialogsAction,
  PrinterSetGroupsAction,
  PrinterSetImgUrlAction,
  PrinterSetModelsAction,
  PrinterSetNewEepromAction,
  PrinterSetPrintqueueAction,
  PrinterSetRecoverAction,
  PrinterSetLastPrintsAction,
  PrinterSetWebcamConfigAction,
  PrinterSortByAction,
  PrinterSortModelsAction,
  PrinterUpdateGCodeAction,
  PrinterUpdateImgUrlsAction,
  PrinterUpdateLogStatusAction,
  PrinterUpdateScriptAction,
  PrinterUpdateStateAction,
  PrinterupdateStatusAction,
  PrinterSetVisibleQuickCommandsAction, PrinterSetVisibleWizardsAction
} from "../actions/PrinterAction";
import {RSUtils} from "../RSUtils";

function sortPrinters(printers: { [key: string]: PrinterState }) {
  const sortedPrinters: string[] = [];
  for (const p in printers) {
    if (printers.hasOwnProperty(p)) {
      sortedPrinters.push(p);
    }
  }
  sortedPrinters.sort(function (aSlug: string, bSlug: string) {
    const a: PrinterState = printers[aSlug];
    const b: PrinterState = printers[bSlug];
    // ['-online', '-active', 'name']);//r;
    if (a.online < b.online) {
      return 1;
    }
    if (a.online > b.online) {
      return -1;
    }
    if (a.active && !b.active) {
      return -1;
    }
    if (!a.active && b.active) {
      return 1;
    }
    const o1 = a.name.toLowerCase();
    const o2 = b.name.toLowerCase();
    if (o1 < o2) {
      return -1;
    }
    if (o1 > o2) {
      return 1;
    }
    return 0;
  });
  return sortedPrinters;
}

function updateActive(printerState: PrintersState): PrintersState {
  if (printerState.sortedPrinters.length === 0) {
    printerState.active = undefined;
    return printerState;
  }
  if (printerState.active === undefined) {
    printerState.active = printerState.sortedPrinters[0];
  }
  if (printerState.autoselectActive && printerState.printers[printerState.active].online !== 1 && printerState.printers[printerState.sortedPrinters[0]].online === 1) {
    printerState.active = printerState.sortedPrinters[0];
  }
  return printerState;
}

const testTempResponse = /\b[TB]\d*:\s*[\d-]/;

function orderBy(list: any[], field: string, reverse: boolean) {
  return list.sort(function (a: any[], b: any[]) {
    // ['-online', '-active', 'name']);//r;
    if (reverse) {
      if (a[field] < b[field]) {
        return -1;
      }
      if (a[field] > b[field]) {
        return 1;
      }
    } else {
      if (a[field] < b[field]) {
        return 1;
      }
      if (a[field] > b[field]) {
        return -1;
      }
    }
    return 0;
  });
}

function sortModels(printerState: PrinterState, allModels: GCodeFileState[]): GCodeFileState[] {
  let models: GCodeFileState[] = [];
  // eslint-disable-next-line
  const reverse = (printerState.sortBy & 1) === 1;
  // eslint-disable-next-line
  switch (printerState.sortBy & 254) {
    case 0:
      models = orderBy(allModels, "name", reverse);
      break;
    case 2:
      models = orderBy(allModels, "created", reverse);
      break;
    case 4:
      models = orderBy(allModels, "printTime", reverse);
      break;
    case 6:
      models = orderBy(allModels, "length", reverse);
      break;
    case 8:
      models = orderBy(allModels, "lines", reverse);
      break;
    case 10:
      models = orderBy(allModels, "filamentTotal", reverse);
      break;
    case 12:
      models = orderBy(allModels, "layer", reverse);
      break;
  }
  if (printerState.modelFilter !== "") {
    const filtered = [];
    for (const v of models) {
      if (v.name.indexOf(printerState.modelFilter) >= 0) {
        filtered.push(v);
      }
    }
    models = filtered;
  }
  if (printerState.activeGroupId !== "*") {
    const filtered = [];
    for (const v of models) {
      if (v.group === printerState.activeGroupId) {
        filtered.push(v);
      }
    }
    models = filtered;
  }
  return models;
}

function appendTemperatureEntry(list: TemperatureDataState[], temp: TemperatureDataState): TemperatureDataState[] {
  let start = 0;
  while (list.length > start && list[start].time + 300000 < temp.time) {
    start++;
  }
  return list.slice(start).concat(temp);
}

function red(state: PrintersState, payload: { slug: string, payload: any }, func: (state: PrinterState, payload: any) => PrinterState) {
  const x = {...state.printers};
  x[payload.slug] = func(state.printers[payload.slug], payload.payload);
  return {...state, printers: x};
}

const reducer = createReducer(INITIAL_PRINTERS_STATE,
  on(PrintersSetActiveAction, (state, payload) =>
    Object.assign({}, state, {active: payload.slug, autoselectActive: false})),
  on(PrintersCreateAction, (state, payload) => {
    const printers: { [key: string]: PrinterState } = Object.assign({}, state.printers);
    printers[payload.slug] =  JSON.parse(JSON.stringify(payload));
    return updateActive(Object.assign({}, state, {printers: printers, sortedPrinters: sortPrinters(printers)}));
  }),
  on(PrintersDestroyAction, (state, payload) => {
    const printers = Object.assign({}, state.printers);
    delete printers[payload.slug];
    return updateActive(Object.assign({}, state, {printers: printers, sortedPrinters: sortPrinters(printers)}));
  }),
  on(PrintersSetAutoSelectAction, (state, payload) => {
    return updateActive(Object.assign({}, state, {autoselectActive: payload.on}));
  }),
  on(PrintersSetStateTimeAction, (state, {stateTime}) =>
    Object.assign({}, state, {stateTime})),

  // Printer related
  on(PrinterUpdateLogStatusAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      return {...state, log: {...state.log, ...payload}};
    })
  ),
  on(PrinterAddLogLineAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const log: LogState = Object.assign({}, state.log, {lastLogStart: payload.id});
      let text = payload.text;
      if (!state.log.logM105) {
        if (text.indexOf("M105") !== -1 || text.indexOf("M409") !== -1 || text.indexOf("M408") !== -1 || testTempResponse.test(text)) {
          text = "";
        }
      }
      if (!state.log.paused && text) {
        const lineText = text.replace(/[<>]/ig, function (m) {
          return "&" + (m === ">" ? "g" : "l") + "t;";
        });
        const newline = payload.time + ": " + lineText;
        if (log.lastText !== lineText) {
          log.history = log.history.slice(0);
          log.history.push("<div class=\"log" + payload.type + "\">" + newline + "</div>");
          log.history.splice(0, log.history.length - log.maxLogLines);
          log.repeat = 1;
          log.lastText = lineText;
        } else { // Repeated lines
          log.repeat++;
          log.history = log.history.slice(0);
          log.history[log.history.length - 1] = "<div class=\"log" + payload.type + "\">" + newline + " (" + log.repeat + ")</div>";
        }
      }
      return Object.assign({}, state, {log: log});
    })
  ),

  on(PrinterSetConfigAction, (outerState, outerPayload) => {
    return red(outerState, outerPayload, (state, payload) => {
      const extra = {config:  JSON.parse(JSON.stringify(payload.config))};
      if(!extra) {
        return state
      }
      if (extra.config.extruders.length > state.activeExtruderId) {
        extra["retractSpeed"] = extra.config.extruders[state.activeExtruderId].retractSpeed;
        extra["extrudeSpeed"] = extra.config.extruders[state.activeExtruderId].extrudeSpeed;
      }
      return {...state, ...extra};
    });
  }),

  on(PrinterSetWebcamConfigAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const extra = {config: Object.assign({}, state.config, {webcams:  JSON.parse(JSON.stringify(payload.webcams))})};
      return Object.assign({}, state, extra);
    })),

  on(PrinterResetImageReferenceAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state) =>
      Object.assign({}, state, {imageUrls: {}}))),

  on(PrinterClearEepromAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state) =>
      Object.assign({}, state, {eeprom: [], eepromErrors: false}))),

  on(PrinterCopyEeepromValueToOrigAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state) => {
      const epr = RSUtils.copyArray(state.eeprom);
      for (const v of epr) {
        v.valueOrig = v.value;
        v.class = "secondary outline";
      }
      return Object.assign({}, state, {eeprom: epr, eepromErrors: false});
    })),

  on(PrinterSetNewEepromAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const epr = RSUtils.copyArray(state.eeprom);
      for (const v of epr) {
        if (v.pos === payload.pos) {
          v.value = payload.value;
        }
      }
      let okAll = true;
      for (const v of epr) {
        const r = /^-?\d+$/;
        const f = /^-?\d+\.?\d*$/;
        let val;
        let ok = true;
        if (v.type === "0") { // byte
          if (r.exec(v.value) === null) {
            ok = false;
          } else {
            val = parseInt(v.value, 10);
            ok = (val > -128 && val < 256);
          }
        }
        if (v.type === "1") { // word
          if (r.exec(v.value) === null) {
            ok = false;
          } else {
            val = parseInt(v.value, 10);
            ok = (val > -32768 && val < 65536);
          }
        }
        if (v.type === "2") { // long
          if (r.exec(v.value) === null) {
            ok = false;
          } else {
            val = parseInt(v.value, 10);
            ok = (val > -2147483648 && val < 4294967296);
          }
        }
        if (v.type === "3") { // float
          if (f.exec(v.value) === null) {
            ok = false;
          }
        }
        if (ok) {
          if (v.value === v.valueOrig) {
            v.class = "secondary outline";
          } else {
            v.class = "success";
          }
        } else {
          v.class = "error";
        }
        okAll = okAll && ok;
      }
      return Object.assign({}, state, {eeprom: epr, eepromErrors: !okAll});
    })),

  on(PrinterAppendEepromAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const epr = RSUtils.copyArray(state.eeprom);
      for (const pl of payload) {
        const text = pl.text;
        let textNU = "";
        let unit = "";
        const p = text.indexOf("[");
        if (p >= 0) {
          textNU = text.substring(0, p);
        } else {
          textNU = text;
        }
        if (p >= 0) {
          const p2 = text.indexOf("]", p);
          if (p2 >= 0) {
            unit = text.substring(p + 1, p2).replace("^2", "²").replace("^3", "³");
          }
        }
        const x = Object.assign({}, pl, {textNU: textNU, unit: unit, class: "secondary outline"});
        epr.push(x);
      }
      return Object.assign({}, state, {eeprom: epr});
    })),

  on(PrinterSetRecoverAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) =>
      Object.assign({}, state, {recover:  JSON.parse(JSON.stringify(payload))}))),
  on(PrinterUpdateScriptAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const extra = {};
      extra[payload.name] = payload.script;
      return Object.assign({}, state, {scripts: Object.assign({}, state.scripts, extra)});
    })),

  on(PrinterSetGroupsAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const newGroups: GroupState[] = [];
      const baseGroups: GroupState[] = [];
      const baseGroups2: GroupState[] = [];
      for (const x in payload.groups) {
        if (payload.groups.hasOwnProperty(x)) {
          const g = payload.groups[x];
          if (g.id === "#" || g.id === "*") {
            baseGroups.push(g);
            if (g.id === "#") {
              baseGroups2.push(g);
            }
          } else {
            newGroups.push(payload.groups[x]);
          }
        }
      }
      return {
        ...state,
        groups: payload.groups,
        activeGroupId: payload.activeGroup,
        groupsSorted: baseGroups.concat(orderBy(newGroups, "name", true)),
        groupsSortedNoAll: baseGroups2.concat(orderBy(newGroups, "name", true))
      };
    })),

  on(PrinterSetDialogsAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      return Object.assign({}, state, {dialogs:  JSON.parse(JSON.stringify(payload))});
    })),

  on(PrinterUpdateStateAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const newState = Object.assign({}, state,  JSON.parse(JSON.stringify(payload.state)));
      newState.activeExtruderId = Math.min(newState.activeExtruderId, newState.extruders.length - 1);
      if (state.activeGroupId !== newState.activeGroupId) {
        newState.modelsFiltered = sortModels(newState, newState.allModels);
      }
      return newState;
    })),

  on(PrinterupdateStatusAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) =>
      Object.assign({}, state,  JSON.parse(JSON.stringify(payload.status))))),

  on(PrinterSetImgUrlAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const a: { [key: string]: string } = {};
      a[payload.key] = payload.url;
      return Object.assign({}, state, {imageUrls: Object.assign({}, state.imageUrls, a)});
    })),

  on(PrinterUpdateImgUrlsAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) =>
      Object.assign({}, state, {imageUrls: Object.assign({}, state.imageUrls,  JSON.parse(JSON.stringify(payload)))}))),

  on(PrinterSetModelsAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {

      return Object.assign({}, state, {
        allModels: payload.slice(0),
        modelsFiltered: sortModels(state, payload.slice(0))
      });
    })),

  on(PrinterSetLastPrintsAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) =>
      Object.assign({}, state, {lastPrints: JSON.parse(JSON.stringify(payload))}))),

  on(PrinterSortByAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) =>
      Object.assign({}, state, {sortBy: payload}))),

  on(PrinterSortModelsAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state) =>
      Object.assign({}, state, {modelsFiltered: sortModels(state, state.allModels.slice(0))}))),

  on(PrinterSetPrintqueueAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) =>
      Object.assign({}, state, {queue:  JSON.parse(JSON.stringify(payload))}))),

  on(PrinterAddBedTemperatureAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const newBed: ToolState[] = RSUtils.copyArray(state.heatedBeds);
      try {
        if (newBed[payload.id]) {
          newBed[payload.id].tempHistory = appendTemperatureEntry(newBed[payload.id].tempHistory, payload.temp);
        }
      } catch (e: any) {
        console.error("PrinterAddBedTemperatureAction", e, newBed, payload);
      }
      return Object.assign({}, state, {heatedBeds: newBed});
    })),

  on(PrinterAddChamberTemperatureAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const newChamber: ToolState[] = RSUtils.copyArray(state.heatedChambers);
      if (newChamber[payload.id]) {
        newChamber[payload.id].tempHistory = appendTemperatureEntry(newChamber[payload.id].tempHistory, payload.temp);
      }
      return Object.assign({}, state, {heatedChambers: newChamber});
    })),

  on(PrinterAddExtruderTemperatureAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const newExtruders: ToolState[] = RSUtils.copyArray(state.extruders);
      if (newExtruders[payload.extruderId]) {
        newExtruders[payload.extruderId].tempHistory = appendTemperatureEntry(newExtruders[payload.extruderId].tempHistory, payload.temp);
      }
      return Object.assign({}, state, {extruders: newExtruders});
    })),

  on(PrinterUpdateGCodeAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      const g: GCodeFileState = payload;
      let idx = -1;
      const newState: PrinterState = Object.assign({}, state);
      if (g.list === "models") {
        for (let i = 0; i < newState.allModels.length; i++) {
          if (newState.allModels[i].id === g.id) {
            idx = i;
            break;
          }
        }
        if (idx !== -1) {
          newState.allModels = newState.allModels.slice(0, idx).concat([g], newState.allModels.slice(idx + 1));
          newState.modelsFiltered = sortModels(newState, newState.allModels);
        }
      } else {
        for (let i = 0; i < newState.queue.length; i++) {
          if (newState.queue[i].id === g.id) {
            idx = i;
            break;
          }
        }
        if (idx !== -1) {
          newState.queue = newState.queue.slice(0, idx).concat([g], newState.queue.slice(idx + 1));
        }
      }
      return newState;
    })),
  on(PrinterSetVisibleQuickCommandsAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      return Object.assign({}, state, {visibleQuickCommands: payload})
    })),
  on(PrinterSetVisibleWizardsAction, (outerState, outerPayload) =>
    red(outerState, outerPayload, (state, payload) => {
      return Object.assign({}, state, {visibleWizards: payload})
    }))
);

export function PrintersReducer(state: PrintersState | undefined, action: Action): PrintersState {
  // state = PrinterReducer(state, action);
  try {
    return reducer(state, action);
  } catch (e) {
    console.error("PrintersReducer error", e);
    return state;
  }
}
