/* (C) 2016-2023 Hot-World GmbH & Co. KG <rlittwin@repetier.com>
 All Rights Reserved

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 It is allowed to use this code to create or use in products connecting to
 Repetier-Server or to use it for own plugins/modules with Repetier-Server.
 This includes free and commercial products.
 Usage for other products is strictly forbidden.
 */

import {Component, Input, ViewChild, ElementRef, OnDestroy, HostBinding, OnChanges, AfterViewInit} from "@angular/core";
import {ConnectionAction, ConnectionHandler, Deferred, RepetierServer, RSUtils} from "binding";
import type {RepetierState, PrinterState, PrintersState, ConnectionState} from "binding";
import {SubscriptionLike} from "rxjs";
import {Store} from "@ngrx/store";

@Component({
  // eslint-disable-next-line @angular-eslint/component-selector
  selector: "rs-websocket-webcam",
  template: `
    <div style="position: relative;height:100%; display: flex" (rsClick)="toggleCam()" (window:resize)="onResize()" >
      <img class="reccam" src="/img/REC.png" *ngIf="active?.rec" title="<?php _('Capturing') ?>" style="position: absolute;width:21px;height:21px;top:5px;left:5px">
      <img #img [src]="url" [style]="myStyle">
    </div>
    `
})
export class RSWebsocketWebcamComponent implements OnDestroy, OnChanges, AfterViewInit {
  @HostBinding("style.position") styPosition = "relative";
  @HostBinding("style.height") styHeight = "100%";
  disposed = false;
  private wasDyn = false;
  url = "/img/nowebcam.png";
  @Input() dynamic = true;

  @Input() imgUrl = '';
  @Input() imgStyle = '';
  @Input() imgClass = '';

  @Input() maxHeight = -1;
  @Input() forceMjpg = false;

  @ViewChild("img", {static: true}) img: ElementRef;
  printers$;
  active: PrinterState = undefined;
  style: any = {};
  lastOrient = -1;
  lastCam = -1;
  webcamId = 0;
  connection: ConnectionState;
  myStyle = '';
  myStyleMargin = '';
  myStyleObjectFit = '';
  myStyleTrans = '';
  myStyleMaxHeight = '';
  myStyleExtra = '';
  urlsWaiting = [];
  rotate = 0;

  viewMode = -1; // -1 = none, 0 = image, 1 = websocket stream, 2 = mjpgStream
  givenUrl = '';
  imgVisibleScroll = true; // is img visible reagrding scroll status
  hidden = '';
  visibilityChange = '';

  init = true;
  ctrl = this;
  ws = null;
  connected = false; // flag if stream is connected, true after establishment
  stopped: any; // Flag if we want stream to exist
  nextTimeout = 100;
  unfreedBlob = undefined; // the blog being shown
  nextBlob = undefined; // read ahead 1 frame
  fpsTimer = null;
  webcamObserver = null;
  resizeObserver = null;
  frames = 0;
  camVisible = false; // for visible observer
  lastSlug = "";

  private subscription: SubscriptionLike;


  constructor(public server: RepetierServer,
              private store: Store<RepetierState>,
              private connectionHandler: ConnectionHandler,
              private elem: ElementRef) {
    this.printers$ = store.select("printers");
    this.subscription = this.printers$.subscribe((p: PrintersState) => {
      if (p.active) {
        this.active = p.printers[p.active];
        if (this.active.slug !== this.lastSlug) {
          this.lastSlug = this.active.slug;
          this.webcamId = -1;
          this.toggleCam();

        }
      }
      this.updateMyStyle();
    });

    if (typeof document.hidden !== "undefined") { // Opera 12.10 and Firefox 18 and later support
      this.hidden = "hidden";
      this.visibilityChange = "visibilitychange";
    }
    // } else if (typeof document.msHidden !== "undefined") {
    //   this.hidden = "msHidden";
    //   this.visibilityChange = "msvisibilitychange";
    // } else if (typeof document.webkitHidden !== "undefined") {
    //   this.hidden = "webkitHidden";
    //   this.visibilityChange = "webkitvisibilitychange";
    // }
  }

  onResize() {
    this.updateMyStyle();
  }

  updateMyStyle () {
    if (!this.img) {
      return;
    }
    const originalWidth = this.img.nativeElement.offsetWidth;
    const originalHeight = this.img.nativeElement.offsetHeight;
    let marginAdd = "0px";
    const marginSub = "0px";
    let imgWidth = originalWidth;
    let imgHeight = originalHeight;
    let scale = 1.0;
    let transY = '0';
    switch (this.ctrl.rotate) {
      case 0:
      case 180:
        imgWidth = originalWidth;
        imgHeight = originalHeight;
        /* if (cam.maxHeight && cam.maxHeight < imgHeight) {
          scale = Math.min(scale, cam.maxHeight / imgHeight)
        } */
        break;
      case 90:
      case 270:
        // noinspection JSSuspiciousNameCombination
        imgWidth = originalHeight;
        // noinspection JSSuspiciousNameCombination
        imgHeight = originalWidth;

        if (this.ctrl.maxHeight > -1) {
          if (this.ctrl.maxHeight && this.ctrl.maxHeight < imgHeight) {
            scale = this.ctrl.maxHeight / imgHeight;
            transY = (-(imgHeight - this.ctrl.maxHeight) / 2).toFixed(0) + 'px';
          }
        }
        marginAdd = Math.round((originalWidth - originalHeight) / 2).toFixed(0) + "px";
        // marginSub = Math.round(-(originalWidth - originalHeight) / 2).toFixed(0) + "px"
        break;
    }
    // console.log(" imgw " + imgWidth + " imgh " + imgHeight, this.rotate);
    if (imgWidth > 0 && imgHeight > 0) {
      const trans = "translateY(" + transY + ") scale(" + scale + ") rotate(" + this.ctrl.rotate + "deg)";
      this.myStyleTrans = `-webkit-transform:${trans};-ms-transform:${trans};transform:${trans};`;
      this.myStyleMargin = `margin:${marginAdd} ${marginSub};`;
      this.myStyleObjectFit = `object-fit:contain; `;
    }

    this.myStyle =  this.myStyleMargin + this.myStyleObjectFit + this.myStyleTrans + this.myStyleMaxHeight + this.myStyleExtra;
    // console.log("updating style", this.myStyle);
  }

  freeBlobs () {
    if (this.unfreedBlob) {
      URL.revokeObjectURL(this.unfreedBlob);
      this.unfreedBlob = null;
    }
    if (this.nextBlob) {
      URL.revokeObjectURL(this.nextBlob);
      this.nextBlob = null;
    }
  }

  startSocket () {
    // console.log('startSocket', this.ws);
    if (this.ws) {
      return;
    }
    if (this.fpsTimer) {
      clearInterval(this.fpsTimer);
    }
    // fpsTimer = setInterval(fpsReporter, 1000)
    // console.log("start socket", this.givenUrl);
    this.camVisible = true;
    this.connected = false;
    this.stopped = false;

    this.ws = new WebSocket(this.givenUrl);
    const ws2 = this.ws; // local to start socket call!
    this.ws.onopen = () => {
      // console.log("onopen", this.ws);
      if (!this.ws) {
        // console.log('closeonopen');
        ws2.close();
        return;
      }
      this.connected = true;
      this.ws.send("next");
      this.ws.send("next");
    };
    this.ws.onerror = (message) => {
      console.log('webcam error', message);
    };
    this.ws.onclose = () => {
      // console.log('webcam closed', this.ws, this, this.stopped, this.connected);
      this.connected = false;
      this.freeBlobs();
      if (this.stopped !== false) {
        // console.log('resolve stop')
        this.stopped.resolve();
        return;
      }
      // console.log('restart connection');
      this.ws = null;
      this.nextTimeout *= 2;
      if (this.nextTimeout > 5000) {
        this.nextTimeout = 5000;
      }
      setTimeout(this.startSocket, this.nextTimeout);
    };
    this.ws.onmessage = (message) => {
      this.nextTimeout = 100;
      // console.log("message:", message);
      if (!this.unfreedBlob) {
        this.unfreedBlob = (window.URL || window.webkitURL).createObjectURL(message.data);
        this.img.nativeElement.src = this.unfreedBlob;
      } else if (!this.nextBlob) {
        this.nextBlob = (window.URL || window.webkitURL).createObjectURL(message.data);
      }
      this.frames++;
    };
  }

  stopSocket () {
    // console.log('stopSocket', this.connected, this.ws);
    if (this.fpsTimer) {
      clearInterval(this.fpsTimer);
    }
    if (this.ws == null) {
      const s = new Deferred();
      s.resolve();
      return s.promise;
    }
    this.stopped = new Deferred();
    this.camVisible = false;
    if (this.connected) {
      this.ws.close();
    }
    this.ws = null;
    return this.stopped.promise;
  }

  // Can be as often as you want also frequently. Will skipped unset
  // values if called too fast

  makeWSURL(): string {
    if (this.active) {
      return (ConnectionHandler.isSecure() ? "wss://" : "ws://") + window.location.host + "/socket/mjpg/" + this.active.slug + "?cam=" + this.webcamId.toString() + "&sess=" + encodeURIComponent(this.connectionHandler.getSession());
    }
     return "";
  }
  assignUrl (url) {
    // console.log("assignurl", url);
    if (url) {
      if (this.urlsWaiting.length > 1) {
        this.urlsWaiting = [this.urlsWaiting[0], url];
      } else {
        this.urlsWaiting.push(url);
      }
    }

    const connectPart2 = () => {
      // console.log('connectPart2', url);
      this.freeBlobs();
      this.givenUrl = url;
      if (url.startsWith('ws')) {
        this.init = false;
        const nativeStream = this.ctrl.forceMjpg || false;
        if (nativeStream) {
          this.viewMode = 2;
          this.img.nativeElement.src = this.givenUrl;
          this.camVisible = true;
          this.stopped = false;
        } else {
          this.viewMode = 1;
          this.imgUrl = this.makeWSURL();
          this.startSocket();
        }
      } else {
        this.stopped = false;
        this.viewMode = 0;
        this.img.nativeElement.src = this.givenUrl;
        this.camVisible = true;
      }
      if (this.urlsWaiting.length) {
        this.assignUrl(false);
      }
    };

    if (this.urlsWaiting.length === 0) {
      return;
    }

    // console.log("urls waiting",this.urlsWaiting);
    url = this.urlsWaiting.shift();
    // console.log("url", url);
    if (url === this.givenUrl) {
      // console.log("is givenUrl", this.givenUrl);
      return;
    }
    if (this.viewMode === 1) {
      this.stopSocket().then(() => {
        connectPart2();
      });
    } else {
      connectPart2();
    }
  }

  handleVisibility () {
    const vis = this.imgVisibleScroll && !document[this.hidden];
    if (this.viewMode < 1 || this.camVisible === vis) {
      return;
    }
    this.camVisible = vis;
    if (this.camVisible) {
      if (this.viewMode === 2) {
        this.img.nativeElement.src = this.givenUrl;
      } else if (this.viewMode === 1) {
        if (this.stopped === false) {
          this.startSocket();
        } else {
          this.stopped.promise.then(() => {
            this.startSocket();
          });
        }
      }
    } else {
      if (this.viewMode === 2) {
        this.img.nativeElement.src = "/img/nowebcam.png";
      } else if (this.viewMode === 1) {
        if (this.stopped === false) {
          this.stopSocket();
        }
      }
    }
  }

  ngOnChanges (diff) {
    // console.log("websocketSlug " + this.imgUrl, diff);
    if (diff.imgUrl && this.viewMode !== -1) {
      // console.log('New image url', this.imgUrl, givenUrl, urlsWaiting)
      this.assignUrl(this.imgUrl);
    }
    if (diff.imgStyle && this.viewMode !== -1) {
      this.ctrl.myStyleExtra = this.imgStyle;
      this.updateMyStyle();
    }
    if (diff.imgClass && this.viewMode !== -1) {
      this.img.nativeElement.className = this.imgClass || '';
    }
    if (diff.maxHeight) {
      this.ctrl.myStyleMaxHeight = `max-height:${this.maxHeight}px`;
      this.updateMyStyle();
    }
  }

  toggleCam() {
    // console.log(this.webcamId);
    this.webcamId++;
    if (this.active.config.webcams) {
      if (this.webcamId >= this.active.config.webcams.length) {
        this.webcamId = 0;
      }
      if (this.active.config.webcams.length > 0) {
        if (this.rotate !== this.active.config.webcams[this.webcamId].orientation) {
          this.rotate = this.active.config.webcams[this.webcamId].orientation;
          this.updateMyStyle();
        }
        this.assignUrl(this.makeWSURL());
      } else {
        this.stopSocket();
        this.freeBlobs();
        this.img.nativeElement.src = "/img/empty.png";
      }

    }
  }

  ngAfterViewInit () {
    // console.log('afterview called');
    this.img.nativeElement.onload = () => {
      if (this.viewMode === 1 && this.ctrl.connected) {
        if (this.unfreedBlob) {
          URL.revokeObjectURL(this.img.nativeElement.src);
          this.unfreedBlob = null;
        }
        if (this.nextBlob) {
          this.unfreedBlob = this.nextBlob;
          this.nextBlob = null;
          this.img.nativeElement.src = this.unfreedBlob;
        } else {
          if (this.ws) {
            this.ws.send("next");
          }
        }
      }
    };
    this.img.nativeElement.style = this.imgStyle + ";object-fit: contain;" || 'object-fit: contain';
    this.img.nativeElement.className = this.imgClass || '';
    this.imgUrl = this.makeWSURL();
    if (this.imgUrl) {
      this.assignUrl(this.imgUrl);
      // console.log(this.imgUrl);
      this.webcamObserver = new IntersectionObserver((entries) => {
        this.imgVisibleScroll = entries[0].isIntersecting;
        this.handleVisibility();
      }, {threshold: [0]});
      this.webcamObserver.observe(this.img.nativeElement);
      this.resizeObserver = new ResizeObserver((elems) => {
        this.updateMyStyle();
      });
      this.resizeObserver.observe(this.img.nativeElement);
      document.addEventListener(this.visibilityChange, this.handleVisibility, false);
      this.updateMyStyle();
    }
  }

  ngOnDestroy () {
    if (this.webcamObserver) {
      this.webcamObserver.disconnect();
    }
    if (this.resizeObserver) {
      this.resizeObserver.disconnect();
    }
    document.removeEventListener(this.visibilityChange, this.handleVisibility, false);
    this.stopSocket();
    this.freeBlobs();
    this.img.nativeElement.src = "/img/empty.png"; // force stop loading mjpg
    this.updateMyStyle();
  }
}
